----------------------------------------------------------------------------------
-- Company: Norges Teknisk Naturvitenskapelige Universitet
-- Engineer: Einar Johan Trøan Sømåen
-- 
-- Module Name:    cell_grid - Behavioral 
-- Project Name: Cellular Automata Accelerator for SHMAC
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use work.CAMComponents.ALL;

--! A collection of cell's that are updated serially.
entity cell_grid is
	GENERIC (num_neighbours : INTEGER := 2; --! The amount of neighbouring cells taken into account when creating generation n+1
			 num_cells : INTEGER := 5 --! The amount of cells grouped together in this grid.
	); -- Simplification to allow some degree of reconfigurability
    Port ( 
			i_rst : IN STD_LOGIC; --! Reset flag, propagated to set reset on rule table 
			i_clk : IN  STD_LOGIC;
			i_start : IN STD_LOGIC;
			i_load_mode : IN STD_LOGIC; --! Load mode flag, if set, then the rule table will run in load mode.
--			iterations_i : IN INTEGER RANGE 0 to 65536;
			i_left_edge : IN STD_LOGIC; --! Input from west cell_grid neighbours
			o_left_edge : OUT STD_LOGIC; --! Output to west cell_grid neighbours
			i_right_edge : IN STD_LOGIC; --! Input from east cell_grid neighbours
			o_right_edge : OUT STD_LOGIC; --! Output to east cell_grid neighbours
			  -- 2D-support:
			o_top : OUT STD_LOGIC_VECTOR(num_cells DOWNTO 1); --! Output to north cell_grid neighbours
			i_top : IN STD_LOGIC_VECTOR(num_cells + 1 DOWNTO 0); --! Input from NW/N/NE cell_grid neighbours
			o_bottom : OUT STD_LOGIC_VECTOR(num_cells DOWNTO 1); --! Output to south cell_grid neighbours
			i_bottom : IN STD_LOGIC_VECTOR(num_cells + 1 DOWNTO 0); --! Input from SW/S/SE cell_grid_neighbours
			-- done-state (Remnants of the chained system, might be usefull again)
--			done_i : IN STD_LOGIC;
--			done_o : OUT STD_LOGIC;
			i_rule_enable_rule_write : IN STD_LOGIC; --! Mask bit to enable ignoring writes to the rule table, thus enabling non-uniformity.
			i_rule_set_rule_mode : IN STD_LOGIC; --! Flag to enable writing the rules defined by i_rule_input_key and i_rule_input_value to the rule table.
			i_rule_input_key : IN STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0); --! The key part of the rule to be written (see rule_table for details)
			i_rule_input_value : IN STD_LOGIC_VECTOR(7 DOWNTO 0); --! The value part of the rule to be written.
			-- Control from grid_controller:
			i_iteration : IN INTEGER RANGE 0 TO 2; --! Control flag, 1 means run, 2 means stall, 0 means "get ready", but has no real purpose at the moment.
			i_update_ctr : IN INTEGER RANGE 0 to num_cells + 1; --! The current cell to be updated
			i_tick : IN STD_LOGIC --! Internal "clock" for the updating of the cells, to determine whether we are in "get rule"-mode, or "write result"-mode.
			);
			
END cell_grid;

ARCHITECTURE Behavioral OF cell_grid IS
	 -- Signals for all the nodes.
	 -- Pedroni p.30
	TYPE t_signal_array IS ARRAY(1 to num_cells) OF STD_LOGIC;
	TYPE t_connection_vector_array IS ARRAY(0 to num_cells + 1) OF STD_LOGIC_VECTOR(num_neighbours DOWNTO 0);
	SIGNAL s_cell_query : t_connection_vector_array;
	-- 2D support
	SIGNAL s_top_neighbours : STD_LOGIC_VECTOR(num_cells + 1 DOWNTO 0);
	SIGNAL s_bottom_neighbours : STD_LOGIC_VECTOR(num_cells + 1 DOWNTO 0);
	-- OTHER
	SIGNAL s_cell_rule_back : t_signal_array;
	SIGNAL s_cell_clks : STD_LOGIC_VECTOR(num_cells DOWNTO 1);

	SIGNAL s_started : STD_LOGIC;
--	SIGNAL done_s : STD_LOGIC;
	SIGNAL s_rules_query : STD_LOGIC_VECTOR(num_neighbours DOWNTO 0); -- TODO: Use me?
	SIGNAL s_rules_result : STD_LOGIC;

	SIGNAL s_rule_input_key : STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0);
	SIGNAL s_rule_input_value : STD_LOGIC_VECTOR(7 DOWNTO 0);
	SIGNAL s_rule_mode : STD_LOGIC_VECTOR(1 DOWNTO 0);
	-- Convention: Higher numbers are WEST/NORTH (W/E has precedence over N/S)
	-- LEFT/RIGHT correspond to WEST/EAST IN same convention.
	-- NW, SW, N, W, c, E, S, NE, SE
	CONSTANT DIR_CENTRE : integer := num_neighbours/2;
	-- N_ is indices into i_neighbour_states, and thus does not have Centre included
	-- These are only really usefull for the rules-setup
	CONSTANT N_DIR_NW : integer := DIR_CENTRE + 3;
	CONSTANT N_DIR_SW : integer := DIR_CENTRE + 2;
	CONSTANT N_DIR_NORTH : integer := DIR_CENTRE + 1;
	CONSTANT N_DIR_WEST : integer := DIR_CENTRE;
	CONSTANT N_DIR_EAST : integer := DIR_CENTRE - 1;
	CONSTANT N_DIR_SOUTH : integer := DIR_CENTRE - 2;
	CONSTANT N_DIR_NE : integer := DIR_CENTRE - 3;
	CONSTANT N_DIR_SE : integer := DIR_CENTRE - 4;
	
begin

	-- 1D: num_neighbours = 2
	-- http://vhdlguru.blogspot.no/2010/09/how-to-use-generate-keyword-for.html
	two_neighbours: IF (num_neighbours = 2) GENERATE
		CELLS: FOR i IN 1 TO num_cells GENERATE BEGIN
			CELL_X : three_cell_ent GENERIC MAP(num_neighbours) PORT MAP( -- LSR
				i_neighbour_states(N2_DIR_WEST) => s_cell_query(i+1)(DIR_CENTRE),
				i_neighbour_states(N2_DIR_EAST) => s_cell_query(i-1)(DIR_CENTRE),
				o_query => s_cell_query(i),
				i_rule_back => s_cell_rule_back(i),
				i_update => s_cell_clks(i),
				i_clk => i_clk
			);
		END GENERATE CELLS;
	END GENERATE two_neighbours;
	
	-- 2D: num_neighbours = 4
	-- http://vhdlguru.blogspot.no/2010/09/how-to-use-generate-keyword-for.html
	four_neighbours: IF (num_neighbours = 4) GENERATE
		CELLS: FOR i IN 1 TO num_cells GENERATE BEGIN
			CELL_X : three_cell_ent GENERIC MAP(num_neighbours) PORT MAP( -- LSR
				i_neighbour_states(N4_DIR_WEST) => s_cell_query(i+1)(DIR_CENTRE),
				i_neighbour_states(N4_DIR_EAST) => s_cell_query(i-1)(DIR_CENTRE),
				i_neighbour_states(N4_DIR_NORTH) => s_top_neighbours(i),
				i_neighbour_states(N4_DIR_SOUTH) => s_bottom_neighbours(i),
				o_query => s_cell_query(i),
				i_rule_back => s_cell_rule_back(i),
				i_update => s_cell_clks(i),
				i_clk => i_clk
		);
		END GENERATE CELLS;
	END GENERATE four_neighbours;
	
	-- 2D: num_neighbours = 8
	-- http://vhdlguru.blogspot.no/2010/09/how-to-use-generate-keyword-for.html
	eight_neighbours: IF (num_neighbours = 8) GENERATE
		CELLS: FOR i IN 1 TO num_cells GENERATE BEGIN
			CELL_X : three_cell_ent GENERIC MAP(num_neighbours) PORT MAP( -- LSR
				i_neighbour_states(N8_DIR_WEST) => s_cell_query(i+1)(DIR_CENTRE),
				i_neighbour_states(N8_DIR_EAST) => s_cell_query(i-1)(DIR_CENTRE),
				i_neighbour_states(N8_DIR_NORTH) => s_top_neighbours(i),
				i_neighbour_states(N8_DIR_NW) => s_top_neighbours(i + 1),
				i_neighbour_states(N8_DIR_NE) => s_top_neighbours(i - 1),
				i_neighbour_states(N8_DIR_SOUTH) => s_bottom_neighbours(i),
				i_neighbour_states(N8_DIR_SW) => s_bottom_neighbours(i + 1),
				i_neighbour_states(N8_DIR_SE) => s_bottom_neighbours(i - 1),
				o_query => s_cell_query(i),
				i_rule_back => s_cell_rule_back(i),
				i_update => s_cell_clks(i),
				i_clk => i_clk
		);
		END GENERATE CELLS;
	END GENERATE eight_neighbours;
	
	-- hard code edge inputs
	edge_query: FOR i IN 0 TO DIR_CENTRE - 1 GENERATE BEGIN
		s_cell_query(0)(i) <= '0';
		s_cell_query(0)(i + DIR_CENTRE + 1) <= '0';
		s_cell_query(num_cells + 1)(i) <= '0';
		s_cell_query(num_cells + 1)(i + DIR_CENTRE + 1) <= '0';
	END GENERATE edge_query;
	
	s_cell_query(0)(DIR_CENTRE) <= i_right_edge;
	s_cell_query(num_cells + 1)(DIR_CENTRE) <= i_left_edge;
	-- hard code right edge to rightmost node
	o_right_edge <= s_cell_query(1)(DIR_CENTRE);
	-- hard code right edge to leftmost node
	o_left_edge <= s_cell_query(num_cells)(DIR_CENTRE);
	
	RULES: rules_ent GENERIC MAP(num_neighbours, DIR_CENTRE, N_DIR_WEST + 1) PORT MAP(
		s_rules_query,
		s_rules_result,
		s_rule_mode,
		i_clk,
		s_rule_input_key,
		s_rule_input_value
	);
	-- These will be ignored if building 1D
	s_top_neighbours <= i_top;
	s_bottom_neighbours <= i_bottom;
	topbot_out: FOR i IN 1 TO num_cells GENERATE BEGIN
		o_bottom(i) <= s_cell_query(i)(DIR_CENTRE);
		o_top(i) <= s_cell_query(i)(DIR_CENTRE);
	END GENERATE;
	
	s_started <= i_start;
		
	PROCESS(i_clk, i_rst) BEGIN
		IF (rising_edge(i_clk)) THEN
			IF (i_rst = '1') THEN
				s_rule_mode <= MODE_RESET;
			-- Only enable writes when the rule write flag is enabled.
			ELSIF (i_rule_set_rule_mode = '1' and i_rule_enable_rule_write = '1') THEN
				s_rule_mode <= MODE_SET_RULE;
			ELSIF (i_load_mode = '1') THEN
				s_rule_mode <= MODE_LOAD;
			ELSE
				s_rule_mode <= MODE_NORMAL;
			END IF;
			IF (num_neighbours /= 2) THEN
				s_rule_input_key <= i_rule_input_key;
			END IF;
			s_rule_input_value <= i_rule_input_value;

		IF (s_started = '1') THEN
			-- Do nothing the first cycle, as we could be triggered
			-- on both a high and low clk, thus making it possible to
			-- have a half-cycle iteration 0, use this for reseting what we need
			-- instead
			IF (i_iteration = 0) THEN
			--	done_s <= '0';
				s_rules_query <= s_cell_query(1);
		--		iteration_count_v := 1; -- we want to start at the high-cycle
			ELSIF (i_iteration = 1) THEN 
				-- On the high part of the clock-cycle, set the rule-query
				IF (i_tick = '1') THEN
					IF (i_update_ctr = num_cells + 1) THEN
						-- Update current state to next state.
						FOR i IN 1 TO num_cells LOOP
							s_cell_clks(i) <= '1';
						END LOOP;
					ELSE
						s_rules_query <= s_cell_query(i_update_ctr);
					END IF;
				ELSE -- tick
				-- On the low part of the clock-cycle, propagate the result.
					IF (i_update_ctr = num_cells + 1) THEN
						-- Update current state to next state.
						FOR i IN 1 TO num_cells LOOP
							s_cell_clks(i) <= '1';
						END LOOP;
					ELSE -- low clock
						-- Must propagate the result now, and not earlier, as
						-- otherwise we will get the old result.
						s_cell_rule_back(i_update_ctr) <= s_rules_result;
						-- Trigger the cell to perform the update
						s_cell_clks(i_update_ctr) <= '0';
					END IF; -- updateCtr
				END IF; -- tick
		--		done_s <= '0';
			ELSE -- iteration-count
		--		done_s <= '1';
			END IF; -- iterations
		ELSE -- started
			s_rules_query <= s_cell_query(1); -- Make sure the rule unit is assigned IN all cases.
	--		done_s <= '1';
		END IF; -- started
		END IF; -- clk
	END PROCESS;
END Behavioral;

