--	Package File Template
--
--	Purpose: This package defines supplemental types, subtypes, 
--		 constants, and functions 


library IEEE;
use IEEE.STD_LOGIC_1164.all;

package CAMComponents is
	component cellacc_toplevel is
		GENERIC (cells_per_sub : INTEGER := 3;
				num_grids : INTEGER := 2;
				num_neighbours : INTEGER := 2;
				height : INTEGER := 2);
		PORT (
			i_start : IN STD_LOGIC;
			i_reset : IN STD_LOGIC;
			i_clk: IN STD_LOGIC;
			i_opcode : IN STD_LOGIC_VECTOR(2 downto 0);
			i_operand : IN STD_LOGIC_VECTOR(28 downto 0);
			i_operand2 : IN STD_LOGIC_VECTOR(height - 1 downto 0);
			o_done: OUT STD_LOGIC;
			o_output : OUT STD_LOGIC_VECTOR(31 downto 0);
			o_output2 : OUT STD_LOGIC_VECTOR(height - 33 downto 0)
		);
	end component;

	COMPONENT cell_grid is
		GENERIC (	num_neighbours : INTEGER;
					num_cells : INTEGER
		); -- Simplification to allow some degree of reconfigurability
		Port ( 
			i_rst : IN STD_LOGIC;
			i_clk : IN  STD_LOGIC;
			i_start : IN STD_LOGIC;
			i_load_mode : IN STD_LOGIC;
			i_left_edge : IN STD_LOGIC;
			o_left_edge : OUT STD_LOGIC;
			i_right_edge : IN STD_LOGIC;
			o_right_edge : OUT STD_LOGIC;
			  -- 2D-support:
			o_top : OUT STD_LOGIC_VECTOR(num_cells downto 1);
			i_top : IN STD_LOGIC_VECTOR(num_cells + 1 downto 0);
			o_bottom : OUT STD_LOGIC_VECTOR(num_cells downto 1);
			i_bottom : IN STD_LOGIC_VECTOR(num_cells + 1 downto 0);
			-- To allow testbenching before synthesizing:
			i_rule_enable_rule_write : IN STD_LOGIC;
			i_rule_set_rule_mode : IN STD_LOGIC;
			i_rule_input_key : IN STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0);
			i_rule_input_value : IN STD_LOGIC_VECTOR(7 downto 0);
			-- Debug:
			i_iteration : IN INTEGER RANGE 0 TO 2;
			i_update_ctr : IN INTEGER RANGE 0 to num_cells + 1;
			i_tick : IN STD_LOGIC);
	END COMPONENT;
	
	COMPONENT grid_controller is
	GENERIC (num_neighbours : INTEGER := 2;
			 num_cells : INTEGER := 5
	); -- Simplification to allow some degree of reconfigurability
    Port (	i_clk : in STD_LOGIC;
			i_iterations : in INTEGER RANGE 0 to 65536;
			o_done : out STD_LOGIC;
			i_start : in STD_LOGIC;
			o_iteration : out INTEGER RANGE 0 to 65536;
			o_start : out STD_LOGIC;
			o_update_ctr : out INTEGER RANGE 0 to num_cells + 1;
			o_tick : out STD_LOGIC
			);
	END COMPONENT;

	COMPONENT three_cell_ent IS
		GENERIC (num_neighbours : INTEGER);
		PORT(
			i_neighbour_states : IN STD_LOGIC_VECTOR(num_neighbours - 1 downto 0);
			o_query : OUT STD_LOGIC_VECtoR(num_neighbours downto 0);
			i_rule_back : IN STD_LOGIC;
			i_update : IN  STD_LOGIC;
			i_clk : IN STD_LOGIC
		);
    END COMPONENT;
	 
	COMPONENT rules_ent IS
		GENERIC (num_neighbours : INTEGER; -- Simplification to allow some degree of reconfigurability
				 center_cell : INTEGER;
				 west_cell_index : INTEGER); -- The cell-position reserved for THIS cell's state
		PORT (
			i_rule_query : IN STD_LOGIC_VECTOR(num_neighbours DOWNTO 0);
			o_output : OUT STD_LOGIC;
			i_mode : IN STD_LOGIC_VECTOR(1 downto 0);
			i_clk : IN STD_LOGIC;
			-- Allows for setting rules
			i_input_key : IN STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0);
			i_input_value : IN STD_LOGIC_VECTOR(7 downto 0)
		);
	 END COMPONENT;

-- Declare constants

	constant OP_HALT : STD_LOGIC_VECTOR := "000"; --! Simply a NOOP
	constant OP_RUN : STD_LOGIC_VECTOR := "001"; --! Run a specified amount of generations
	constant OP_SETMASK : STD_LOGIC_VECTOR := "010"; --! Set a mask specifying which cell_grids write any new rules
	constant OP_GETINFO : STD_LOGIC_VECTOR := "011"; --! Query information about the accelerator (sizes and neighbourhood)
	constant OP_SETRULE : STD_LOGIC_VECTOR := "100"; --! Store a rule in the active rule tables
	constant OP_SETEDGE : STD_LOGIC_VECTOR := "101"; --! Enable/disable edge-wrapping/set static edge value.
	constant OP_LOADCOL : STD_LOGIC_VECTOR := "110"; --! Load one or more columns into the grid
	constant OP_RST : STD_LOGIC_VECTOR := "111"; --! Reset the grid to 0 or 1.
	
	-- Rule table modes
	constant MODE_LOAD : STD_LOGIC_VECTOR := "10"; --! A right shift rule, to allow for loading columns into the CA.
	constant MODE_SET_RULE : STD_LOGIC_VECTOR := "01"; --! Write mode for the rule tables.
	constant MODE_RESET : STD_LOGIC_VECTOR := "11"; --! Reset the entire rule table to 0 or 1.
	constant MODE_NORMAL : STD_LOGIC_VECTOR := "00"; --! Normal (read only) operation.
	
	-- Convention: Higher numbers are WEST/NORTH (W/E has precedence over N/S)
	-- LEFT/RIGHT correspond to WEST/EAST in same convention.
	-- NW, SW, N, W, c, E, S, NE, SE
	-- N_ is indices into NEIGHBOUR_STATES, and thus does not have Centre included
	-- 2 neighbours (high to low, BE) 
	-- W, c, E
	constant N2_DIR_WEST : integer := 1;
	constant N2_DIR_EAST : integer := 0;
	-- 4 neighbours (high to low, BE) 
	-- N, W, c, E, S
	constant N4_DIR_NORTH : integer := 3;
	constant N4_DIR_WEST : integer := 2;
	constant N4_DIR_EAST : integer := 1;
	constant N4_DIR_SOUTH : integer := 0;
	-- 8 neighbours (high to low, BE) 
	-- NE, SE, N, E, c, W, S, NW, SW
	constant N8_DIR_NW : integer := 7;
	constant N8_DIR_SW : integer := 6;
	constant N8_DIR_NORTH : integer := 5;
	constant N8_DIR_WEST : integer := 4;
	constant N8_DIR_EAST : integer := 3;
	constant N8_DIR_SOUTH : integer := 2;
	constant N8_DIR_NE : integer := 1;
	constant N8_DIR_SE : integer := 0;	
 
	 
end CAMComponents;
