----------------------------------------------------------------------------------
-- Company: Norges Teknisk Naturvitenskapelige Universitet
-- Engineer: Einar Johan Trøan Sømåen
-- 
-- Module Name:    GridController - Behavioral 
-- Project Name: Cellular Automata Accelerator for SHMAC
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- The GridController is basically the repeated counting-logic of the various
-- CellGrids, the logic in all the CellGrids is equal, so this reduces the amount
-- of required counting in the circuit.
entity grid_controller is
	GENERIC (num_neighbours : INTEGER := 2;
			 num_cells : INTEGER := 5
	); -- Simplification to allow some degree of reconfigurability
    Port (	i_clk : IN STD_LOGIC;
			i_iterations : IN  INTEGER RANGE 0 to 536870911; -- 2^29 - 1
			o_done : OUT STD_LOGIC;
			i_start : IN STD_LOGIC;
			o_iteration : OUT INTEGER RANGE 0 to 2;
			o_start : OUT STD_LOGIC;
			o_update_ctr : OUT INTEGER RANGE 0 to num_cells + 1;
			o_tick : OUT STD_LOGIC
			);
end grid_controller;

architecture Behavioral of grid_controller is
	-- Purely intended for forcing ModelSim to print every generation, even when they're exactly equal
	-- SIGNAL dbg_generation_count : INTEGER RANGE 0 TO 526870911;
begin
	o_start <= i_start;
	process(i_clk) 
		variable v_iteration_count : INTEGER RANGE 0 TO 526870911; -- 2^29 - 1
		variable v_iteration_output : INTEGER RANGE 0 TO 2;
		variable v_update_ctr : INTEGER RANGE 0 to num_cells + 1;
		variable v_tick : STD_LOGIC;
	begin
	if (rising_edge(i_clk)) then
		if (i_start = '1') then
			-- Do nothing the first cycle, as we could be triggered
			-- on both a high and low clk, thus making it possible to
			-- have a half-cycle iteration 0, use this for reseting what we need
			-- instead
			if (v_iteration_count = 0 and i_iterations > 0) then
				o_done <= '0';
				v_tick := '1';
				v_update_ctr := 1;
				v_iteration_count := 1; -- we want to start at the high-cycle
				v_iteration_output := 1;
			elsif (i_iterations > 0 and v_iteration_count <= i_iterations) then 
				v_iteration_output := 1;
				-- On the high part of the clock-cycle, set the rule-query
				if (v_tick = '1') then
					v_tick := '0';
				else -- tick
				-- On the low part of the clock-cycle, propagate the result.
					if (v_update_ctr = num_cells + 1) then
						v_update_ctr := 1;
						v_iteration_count := v_iteration_count + 1;
					else -- low clock
						v_update_ctr := v_update_ctr + 1;
					end if; -- updateCtr
					v_tick := '1';
				end if; -- tick
				o_done <= '0';
			else -- iteration-count
				v_iteration_output := 2;
				o_done <= '1';
			end if; -- iterations
		else -- started
			v_iteration_output := 2;
			v_iteration_count := 0;
			v_update_ctr := 1;
			o_done <= '1';
		end if; -- started
		o_update_ctr <= v_update_ctr;
		o_iteration <= v_iteration_output;
		o_tick <= v_tick;
		--Uncomment this to be able to print every gen in ModelSim
		--dbg_generation_count <= v_iteration_count;
	end if; -- clk
	end PROCESS;
end Behavioral;

