----------------------------------------------------------------------------------
-- Company: Norges Teknisk Naturvitenskapelige Universitet
-- Engineer: Einar Johan Trøan Sømåen
-- 
-- Module Name:  rule_table - Behavioral 
-- Project Name: Cellular Automata Accelerator for SHMAC
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;
LIBRARY std;
USE std.standard.all;


ENTITY rules_ent IS
	GENERIC (	num_neighbours : INTEGER := 4; --! The amount of neighbours considered when deciding generation n+1
				center_cell : INTEGER := 1; --! Index of the central cell (i.e. the cell whos neighbours we are looking at)
				west_cell_index : INTEGER := 0); --! Index of the cell that is directly WEST of center (used for load-mode).
	PORT (
		i_rule_query : IN STD_LOGIC_VECTOR(num_neighbours DOWNTO 0); --! The input from a cell looking for the corresponding next state.
		o_output : OUT STD_LOGIC; --! The result from a rule lookup
--	 	RST : IN STD_LOGIC;
		i_mode : IN STD_LOGIC_VECTOR(1 downto 0); --! The mode that the table is to work in, can be LOAD/SET_RULE/RESET/NORMAL
		i_clk : IN STD_LOGIC;
		--! The key is used for storing rules to the table, and is simply
		--! any bits above the 3 LSB bits.
		--! Where a full rule is accessed with (num_neighbours + 1) bits
		--! when reading, we use 3 fewer bits when storing. Reducing the
		--! amount of writes needed from (num_neighbours + 1)^2 to
		--! (num_neighbours - 2)^2. The value part is then the 8 values
		--! corresponding to each value between the key and the next key.
		--! This means in effect that the value is a Wolfram-number,
		--! and the key is used as a prefix to that specific number. 
		i_input_key : IN STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0); 
		--! The value part of a rule, always 8 bits long. (Corresponds to a Wolfram number)
		i_input_value : IN STD_LOGIC_VECTOR(7 downto 0)	
	);
END rules_ent;

ARCHITECTURE rules OF rules_ent IS
	CONSTANT MODE_LOAD : STD_LOGIC_VECTOR := "10";
	CONSTANT MODE_SET_RULE : STD_LOGIC_VECTOR := "01";
	CONSTANT MODE_RESET : STD_LOGIC_VECTOR := "11";
	CONSTANT MODE_NORMAL : STD_LOGIC_VECTOR := "00";

	SIGNAL s_table : STD_LOGIC_VECTOR(2**(num_neighbours + 1) - 1 DOWNTO 0);
BEGIN

	WITH i_mode SELECT 
		o_output <= 	i_rule_query(center_cell) WHEN MODE_SET_RULE,
						i_rule_query(west_cell_index) WHEN MODE_LOAD,
						s_table(conv_integer(i_rule_query)) WHEN MODE_NORMAL,
						i_input_value(0) WHEN OTHERS;

  set_rules : PROCESS(i_clk) 
	variable v_input_key_extended : STD_LOGIC_VECTOR(num_neighbours DOWNTO 0);
  BEGIN
	IF (rising_edge(i_clk)) THEN
		IF (i_mode = MODE_RESET) THEN
			s_table(2**(num_neighbours + 1) - 1 DOWNTO 0) <= (OTHERS => i_input_value(0));
		ELSIF (i_mode = MODE_SET_RULE) THEN
			IF (num_neighbours = 2) THEN
				s_table <= i_input_value;
			ELSE
				v_input_key_extended(num_neighbours downto 3) := i_input_key; -- Has to be XXX YYY 111, since we get the data in in high to low form.
				v_input_key_extended(2 downto 0) := "111";
				s_table(conv_integer(v_input_key_extended) downto conv_integer(v_input_key_extended - "111")) <= i_input_value;
			END IF;
		END IF;
	end if; -- clk
  END PROCESS;
END rules;
