----------------------------------------------------------------------------------
-- Company: Norges Teknisk Naturvitenskapelige Universitet
-- Engineer: Einar Johan Tran Smen
-- 
-- Module Name:  toplevel - Behavioral 
-- Project Name: Cellular Automata Accelerator for SHMAC
----------------------------------------------------------------------------------
LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE ieee.std_logic_unsigned.all;
USE work.CAMComponents.all;
USE ieee.std_logic_arith.all;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

--! @brief The toplevel component of the Cellular Automata Accelerator
--! @details This is the component responsible for handling reactions to outside
--! stimuli, and also providing output to the outside. Mainly this is done
--! through the opcode/operand ports, and all output comes through the single
--! output-port. Usage is a matter of providing the necessary opcodes in sequence,
--! and listening on the output-port when the done-flag is high.
ENTITY cellacc_toplevel IS
	GENERIC (cells_per_sub : INTEGER := 3;
				num_grids : INTEGER := 2;
				num_neighbours : INTEGER := 2;
				height : INTEGER := 2);
	PORT (
		i_start : IN STD_LOGIC; --! Start flag, driven from the IFM.
		i_reset : IN STD_LOGIC; --! Reset flag, driven from the IFM
		i_clk : IN STD_LOGIC;
		i_opcode : IN STD_LOGIC_VECTOR(2 DOWNTO 0); --! Opcode, decides which operation is to be performed
		i_operand : IN STD_LOGIC_VECTOR(28 DOWNTO 0); --! Operand, any arguments that follows the given opcode.
		i_operand2 : IN STD_LOGIC_VECTOR(height - 1 downto 0); --! Additional operand-bits for heights above 29.
		o_done : OUT STD_LOGIC; --! States whether the accelerator is ready for a new operation
		o_output : OUT STD_LOGIC_VECTOR(31 DOWNTO 0); --! The accelerators output, either the easternmost column, or grid-information, depending on operation
		o_output2 : OUT STD_LOGIC_VECTOR(height - 33 downto 0) --! Additional outputs beyond the first 32.
	);
END cellacc_toplevel;

ARCHITECTURE behavioral OF cellacc_toplevel IS

	TYPE t_vector_array IS ARRAY(0 to num_grids + 1) of STD_LOGIC_VECTOR(1 DOWNTO 0);
	TYPE t_inter_signal_array IS ARRAY(0 to height) of STD_LOGIC_VECTOR(num_grids DOWNTO 0);
	TYPE t_rule_mask_array IS ARRAY(0 to height - 1) of STD_LOGIC_VECTOR(num_grids - 1 DOWNTO 0);
	TYPE t_signal_array IS ARRAY(0 to num_grids - 1) of STD_LOGIC;
	TYPE t_downwards_connected_array IS ARRAY(0 to height) of STD_LOGIC_VECTOR(cells_per_sub * num_grids + 1 DOWNTO 0); -- TODO: Fix the corners
	
	SIGNAL s_rule_mode_write_enable : t_rule_mask_array;
	SIGNAL s_grid_interconnect_N2S : t_downwards_connected_array;
	SIGNAL s_grid_interconnect_S2N : t_downwards_connected_array;
	-- Convention: 0 is the right edge, max is left edge
	SIGNAL s_grid_interconnect_L2R : t_inter_signal_array;
	SIGNAL s_grid_interconnect_R2L : t_inter_signal_array;

	SIGNAL s_rst : STD_LOGIC;
	SIGNAL s_load_mode : STD_LOGIC;
	SIGNAL s_iterations : INTEGER RANGE 0 to 536870911; -- 2^29 - 1

	SIGNAL s_load_edge_in : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
	SIGNAL s_left_edge_in : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
	SIGNAL s_left_edge_out : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
	SIGNAL s_right_edge_in : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
	SIGNAL s_right_edge_out : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
	-- 2D-support
	--SIGNAL topOut_s : STD_LOGIC_VECTOR(cells_per_sub DOWNTO 0);
	SIGNAL s_top_in : STD_LOGIC_VECTOR(cells_per_sub + 1 DOWNTO 0);
	--SIGNAL bottomOut_s : STD_LOGIC_VECTOR(cells_per_sub DOWNTO 0);
	SIGNAL s_bottom_in : STD_LOGIC_VECTOR(cells_per_sub + 1 DOWNTO 0);
	-- other logic
	SIGNAL s_done : STD_LOGIC;
	SIGNAL s_set_rule_mode : STD_LOGIC;
	SIGNAL s_reg_input_key : STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0);
	SIGNAL s_reg_input_val : STD_LOGIC_VECTOR(7 DOWNTO 0);

	-- Control E/W-wrapmux
	SIGNAL s_wrap_EW_enabled : STD_LOGIC;
	SIGNAL s_wrap_NS_enabled : STD_LOGIC;
	
	signal s_current_op : STD_LOGIC_VECTOR(31 DOWNTO 0);

	-- Information output;
	SIGNAL s_write_information : STD_LOGIC;
	SIGNAL s_grid_information : STD_LOGIC_VECTOR(31 DOWNTO 0);
	
	-- GridController
	SIGNAL s_iteration_for_grid : INTEGER RANGE 0 to 65536;
	SIGNAL s_controller_done : STD_LOGIC;
	SIGNAL s_controller_start : STD_LOGIC;
	SIGNAL s_controller_start_out : STD_LOGIC;
	SIGNAL s_update_counter : INTEGER RANGE 0 to cells_per_sub + 1;
	SIGNAL s_tick : STD_LOGIC;
begin
	CONTROLLER : grid_controller GENERIC MAP(num_neighbours, cells_per_sub) PORT MAP (
		i_clk => i_clk,
		i_iterations => s_iterations,
		o_done => s_controller_done,
		i_start => s_controller_start,
		o_start => s_controller_start_out,
		o_iteration => s_iteration_for_grid,
		o_update_ctr => s_update_counter,
		o_tick => s_tick
	);
	
	
	-- Grid coords: origo is SE.
	YGRIDS : FOR y IN 0 TO height - 1 GENERATE BEGIN -- TODO: Fix Y-coordinates
		GRIDS: FOR x IN 0 TO num_grids - 1 GENERATE BEGIN
			GRID: cell_grid GENERIC MAP(num_neighbours, cells_per_sub) PORT MAP(
			i_rst => s_rst,
			i_clk => i_clk, -- TODO: Clock-gating.
			i_start => s_controller_start_out,
			i_load_mode => s_load_mode,
			i_left_edge => s_grid_interconnect_L2R(y)(x + 1),
			o_left_edge => s_grid_interconnect_R2L(y)(x + 1),
			i_right_edge => s_grid_interconnect_R2L(y)(x),
			o_right_edge => s_grid_interconnect_L2R(y)(x),
			-- 2D support (TODO, fix coords)
			o_top => s_grid_interconnect_S2N(y + 1)((x + 1) * cells_per_sub DOWNTO (x) * cells_per_sub + 1),
			i_top => s_grid_interconnect_N2S(y + 1)((x + 1) * cells_per_sub + 1 DOWNTO (x) * cells_per_sub),
			o_bottom => s_grid_interconnect_N2S(y)((x + 1) * cells_per_sub DOWNTO (x) * cells_per_sub + 1),
			i_bottom => s_grid_interconnect_S2N(y)((x + 1) * cells_per_sub + 1 DOWNTO (x) * cells_per_sub),
			-- other logic
			i_rule_enable_rule_write => s_rule_mode_write_enable(y)(x),
			i_rule_set_rule_mode => s_set_rule_mode,
			i_rule_input_key => s_reg_input_key,
			i_rule_input_value => s_reg_input_val,
			i_iteration => s_iteration_for_grid,
			i_update_ctr => s_update_counter,
			i_tick => s_tick
		);
		END GENERATE GRIDS;
		-- Connect the interconnect of this row to the edge
		s_grid_interconnect_L2R(y)(num_grids) <= s_left_edge_in(y); -- Left IN
		s_left_edge_out(y) <= s_grid_interconnect_R2L(y)(num_grids); -- Left OUT
		s_right_edge_out(y) <= s_grid_interconnect_L2R(y)(0); -- Right IN
		s_grid_interconnect_R2L(y)(0) <= s_right_edge_in(y); -- Right OUT
	END GENERATE YGRIDS;
	-- doneGrids(height - 1)(num_grids) is the final done-signal
--	doneGrids_s(0)(0) <= '1';
	
	-- Set NW,NE,SW,SE interconnects TODO: Combine with southwards
	north_corners: FOR i IN 1 TO height - 1 GENERATE BEGIN
		s_grid_interconnect_N2S(i)(0) <= s_right_edge_in(i); -- NE
		s_grid_interconnect_N2S(i)(cells_per_sub * num_grids + 1) <= s_left_edge_in(i); -- NW
	END GENERATE north_corners;

	south_corners: FOR i IN 1 TO height - 1 GENERATE BEGIN
		s_grid_interconnect_S2N(i)(0) <= s_right_edge_in(i - 1); -- SE
		s_grid_interconnect_S2N(i)(cells_per_sub * num_grids + 1) <= s_left_edge_in(i - 1); -- SW
	END GENERATE south_corners;
	
	-- Wrap corners
	--NW Corner:  
	s_grid_interconnect_S2N(height)(cells_per_sub * num_grids + 1) <= s_grid_interconnect_S2N(height)(1) 
		WHEN (s_wrap_NS_enabled = '1' and s_wrap_EW_enabled = '1') 
		ELSE '0';
	--NE Corner: (Wrapping dictates we want NW-cell here, as this is input in SE)
	s_grid_interconnect_S2N(height)(0) <= s_grid_interconnect_S2N(height)(cells_per_sub * num_grids) WHEN (s_wrap_NS_enabled = '1' and s_wrap_EW_enabled = '1') ELSE '0';
	--SW Corner:
	s_grid_interconnect_N2S(0)(cells_per_sub * num_grids + 1) <=  s_grid_interconnect_N2S(0)(1) WHEN (s_wrap_NS_enabled = '1' and s_wrap_EW_enabled = '1') ELSE '0';
	--SE Corner: (Wrapping dictates we want SW-cell here, as this is input in NE)
	s_grid_interconnect_N2S(0)(0) <= s_grid_interconnect_N2S(0)(cells_per_sub * num_grids) WHEN (s_wrap_NS_enabled = '1' and s_wrap_EW_enabled = '1') ELSE '0';

	-- Wrap N/S
	s_grid_interconnect_N2S(height)(cells_per_sub * num_grids + 1 DOWNTO 0) <= s_grid_interconnect_N2S(0)(cells_per_sub * num_grids + 1 DOWNTO 0) WHEN (s_wrap_NS_enabled = '1') ELSE
																		(OTHERS => '0'); -- TOP connected to BOTTOM
	s_grid_interconnect_S2N(0)(cells_per_sub * num_grids + 1 DOWNTO 0) <= s_grid_interconnect_S2N(height)(cells_per_sub * num_grids + 1 DOWNTO 0) WHEN (s_wrap_NS_enabled = '1') ELSE
																		(OTHERS => '0'); -- BOTTOM connected to TOP
	-- Wrap E/W
	s_left_edge_in <= s_load_edge_in WHEN (s_wrap_EW_enabled = '0') ELSE
					s_right_edge_out WHEN (s_wrap_EW_enabled = '1') ELSE
					(OTHERS=>'0');
					
	s_right_edge_in <= s_left_edge_out WHEN (s_wrap_EW_enabled = '1') ELSE
					(OTHERS=>'0');

	handle_clk: PROCESS(i_clk, i_start)
		VARIABLE v_op : STD_LOGIC_VECTOR(2 DOWNTO 0);
		VARIABLE v_full_op : STD_LOGIC_VECTOR(31 DOWNTO 0);
		VARIABLE v_rst : STD_LOGIC := '1';
		VARIABLE v_iterations : INTEGER RANGE 0 TO 536870911; -- 2^29 - 1
		VARIABLE v_reg_input_val : STD_LOGIC_VECTOR(7 DOWNTO 0);
		VARIABLE v_reg_input_key : STD_LOGIC_VECTOR(num_neighbours - 3 DOWNTO 0);
		VARIABLE v_start : STD_LOGIC;
		VARIABLE v_load_mode : STD_LOGIC;
		VARIABLE v_set_rule_mode : STD_LOGIC := '0';
		VARIABLE v_done : STD_LOGIC := '0';
		VARIABLE v_wait_a_tick : STD_LOGIC := '0';
		VARIABLE v_write_signals : STD_LOGIC := '0';
		VARIABLE v_wrap_north_south : STD_LOGIC := '1';
		VARIABLE v_wrap_east_west : STD_LOGIC := '0';
		VARIABLE v_load_edge_in : STD_LOGIC_VECTOR(height - 1 DOWNTO 0);
		-- For non-uniformity, as well as avoiding the rule tables from being optimized into one
		VARIABLE v_clear_all_write_masks : STD_LOGIC := '0';
		VARIABLE v_write_mask : STD_LOGIC := '0';
		VARIABLE v_write_mask_x : INTEGER RANGE 0 TO 511;
		VARIABLE v_write_mask_y : INTEGER RANGE 0 TO 511;
		VARIABLE v_write_mask_full_row : STD_LOGIC;
		VARIABLE v_write_mask_full_col : STD_LOGIC;
		VARIABLE v_write_mask_value : STD_LOGIC;
		-- For getting information
		VARIABLE v_write_information : STD_LOGIC;
	BEGIN
		IF (rising_edge(i_clk)) THEN
			-- Ignore the chained done-flags for now, and use the gridController's output instead.
			v_done := s_controller_done;--doneGrids_s(height - 1)(num_grids);
			v_write_signals := '0';
			v_clear_all_write_masks := '0';
			v_write_mask := '0';
			-- Handle done, by stopping the grid when it is done?

			-- Overwrite the opcode when reset is high.
			IF (i_reset = '1') THEN
				v_op := OP_RST;
				-- Need to explicitly handle this, as we can't rely on the operand
				-- field at this point.
				v_full_op := (OTHERS => '0');
				v_full_op(31 DOWNTO 29) := v_op;
			ELSE
				v_op := i_opcode;
				-- Reconstruct the full opcode for comparison.
				v_full_op := v_op & i_operand;
			END IF;


			v_start := s_controller_start;

			IF (v_wait_a_tick = '1') THEN -- Opcodes take a tick (clock cycle) to take effect
				v_wait_a_tick := '0';
				-- Preserve current state.
				v_done := '0';
				v_write_signals := '1';
				v_write_information := s_write_information;
				s_current_op <= v_full_op; -- HUH
			ELSIF (v_done = '1') THEN -- Only handle other ops when the grid is done
				-- Handle other ops
				v_rst := '0';
				v_set_rule_mode := '0';
				v_load_mode := '0';
				v_start := '0';
				v_iterations := 0;

				-- Only perform when start or reset-flag is high
				IF (i_start = '1' OR i_reset = '1') THEN
					v_write_information := '0';
					v_wait_a_tick := '1';
					IF (v_op = OP_RST) THEN
						v_rst := '1';
						v_iterations := 1; -- We need one iteration to perform a full reset
						v_reg_input_val := (OTHERS => v_full_op(28)); -- pick the bit to reset everything to
						IF (num_neighbours /= 2) THEN
							v_reg_input_key := (OTHERS => '0'); -- Just assure that the input-key is set.
						END IF;
						v_start := '1';
						s_current_op <= v_full_op;
						v_wait_a_tick := '1';
						-- Can't pull down done early here, as it would override the iteration done.
						-- We do not care about load_mode and set_rule_mode here, but set them anyhow:
						v_load_mode := '0';
						v_set_rule_mode := '0';
						v_wrap_east_west := '0';
						v_wrap_north_south := '0';
						-- Apply signals this tick
						v_write_signals := '1';
						-- Enable writes to all rule tables.
						v_clear_all_write_masks := '1';
						v_write_mask_value := '1';
						-- Not sure about these:
						-- Need to use fullOp_v, as this op can be triggered by RESET
						v_load_edge_in(height - 1 DOWNTO 0) := (OTHERS => v_full_op(28));
						s_grid_information(31 DOWNTO 0) <= (OTHERS => '0');
					ELSIF (v_op = OP_RUN) THEN
						v_iterations := conv_integer(i_operand(28 DOWNTO 0));
						v_start := '1';
						v_done := '0';
					ELSIF (v_op = OP_HALT) THEN
						v_iterations := 0;
						v_done := '0';
					ELSIF (v_op = OP_SETEDGE) THEN
						IF (i_operand(25) = '1') THEN
							v_wrap_east_west := '1';
						ELSE
							v_load_edge_in(height - 1 DOWNTO 0) := i_operand2(height - 1 DOWNTO 0);
							v_wrap_east_west := '0';
						END IF;
						v_wrap_north_south := i_operand(24);
						v_iterations := 0;
						v_done := '0'; -- Hold Done low for a clock cycle
					ELSIF (v_op = OP_SETMASK) THEN
						v_write_mask_full_row := i_operand(28);
						v_write_mask_full_col := i_operand(27);
						v_write_mask_value := i_operand(26);
						v_write_mask_x := conv_integer(i_operand(25 DOWNTO 17));
						v_write_mask_y := conv_integer(i_operand(16 DOWNTO 8));
						v_write_mask := '1';
					ELSIF (v_op = OP_GETINFO) THEN
						v_write_information := '1';
						-- Get width/height
						IF (i_operand(28) = '1') THEN
							s_grid_information(31 DOWNTO 16) <= conv_std_logic_vector(num_grids, 16);
							s_grid_information(15 DOWNTO 0) <= conv_std_logic_vector(height, 16);
						ELSE
							s_grid_information(31 DOWNTO 16) <= conv_std_logic_vector(cells_per_sub, 16);
							s_grid_information(15 DOWNTO 0) <= conv_std_logic_vector(num_neighbours, 16);
						END IF;
						v_done := '0';
					ELSIF (v_op = OP_LOADCOL) THEN -- Side-effect: leftEdgeIn is stuck
						v_load_mode := '1';
						v_iterations := conv_integer(i_operand(28 DOWNTO 0));
						v_load_edge_in(height - 1 DOWNTO 0) := i_operand2(height - 1 DOWNTO 0);	
						v_start := '1';
						v_done := '0';
					ELSIF (v_op = OP_SETRULE) THEN
						v_set_rule_mode := '1';
						IF (num_neighbours /= 2) THEN
							v_reg_input_key := i_operand(20 DOWNTO 20 - (num_neighbours - 3));
						END IF;
						v_reg_input_val := i_operand(28 downto 21);
						v_done := '0'; -- Hold Done low for a clock cycle
						v_wait_a_tick := '1';
					END IF; -- OP_RUN
					s_current_op <= v_full_op;
				END IF; -- i_start
				v_write_signals := '1';
			END IF; -- v_done
			s_wrap_EW_enabled <= v_wrap_east_west;
			s_wrap_NS_enabled <= v_wrap_north_south;
			s_load_edge_in <= v_load_edge_in;
			IF (v_write_signals = '1') THEN
				IF (v_write_mask = '1' and v_write_mask_full_row = '1' and v_write_mask_full_col = '1') THEN
					v_clear_all_write_masks := '1';
				END IF;
				IF (v_clear_all_write_masks = '1') THEN
					FOR y IN 0 TO height - 1 LOOP
						FOR x IN 0 TO num_grids - 1 loop
							s_rule_mode_write_enable(y)(x) <= v_write_mask_value;
						END LOOP;
					END LOOP;
					v_clear_all_write_masks := '0';
				END IF;
				IF (v_write_mask = '1') THEN
					IF (v_clear_all_write_masks = '0') THEN
						-- At this point we've handled full-row = 1 AND full-row = 1
						-- so here we only need to handle 3 cases:
						-- Full row only
						-- Full col only
						-- Neither (single mask value)
						IF (v_write_mask_full_row = '1') THEN
							IF (v_write_mask_Y < height) THEN
								FOR x IN 0 TO num_grids - 1 LOOP
									s_rule_mode_write_enable(v_write_mask_Y)(x) <= v_write_mask_value;
								END LOOP;
							END IF;
						ELSIF (v_write_mask_full_col = '1') THEN
							IF (v_write_mask_X < num_grids) THEN
								FOR y IN 0 TO height - 1 LOOP
									s_rule_mode_write_enable(y)(v_write_mask_X) <= v_write_mask_value;
								END LOOP;
							END IF;
						ELSE -- Single mask value
							IF (v_write_mask_X < num_grids and v_write_mask_Y < height) THEN
								s_rule_mode_write_enable(v_write_mask_Y)(v_write_mask_X) <= v_write_mask_value;
							END IF;
						END IF;
					END IF; -- clearAllWriteMasks_v
				END IF; -- writeMasks_v
				s_rst <= v_rst;
				s_controller_start <= v_start;
				s_iterations <= v_iterations;
				s_reg_input_val <= v_reg_input_val;
				IF (num_neighbours /= 2) THEN
					s_reg_input_key <= v_reg_input_key;
				END IF;
				s_controller_start <= v_start;
				s_load_mode <= v_load_mode;
				s_set_rule_mode <= v_set_rule_mode;
				s_write_information <= v_write_information;
			END IF; -- v_write_signals
			o_done <= v_done;
			IF (v_done = '1') THEN
				IF (s_write_information = '1') THEN
					o_output <= s_grid_information;
					IF (height > 32) THEN
						o_output2 <= (OTHERS => '0');
					END IF;
				ELSE
					IF (height - 1 < 31) THEN
						o_output(31 DOWNTO height) <= (OTHERS => '0');
					END IF;
					FOR i IN 0 TO height - 1 LOOP
						IF (i < 32) THEN
							-- Handle the lower 32 bits of output
							o_output(i) <= s_grid_interconnect_L2R(i)(0);
						ELSE
							-- Handle the case with more than 32 bits of output
							-- Lowest 32 bits of output go to o_output1, filling from 0 to 31
							-- When we have more than 31, then bit 32 and up go to bit 0 and up in o_output2.
							o_output2(i - 32) <= s_grid_interconnect_L2R(i)(0);
						END IF; -- i < 32
					END LOOP;
				END IF; -- s_write_information
			END IF;
		END IF; -- clk
	END PROCESS handle_clk;
END Behavioral;

