#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "cam.h"

static unsigned int g_num_neighbours;
static unsigned int g_grid_height;
static unsigned int g_grid_width;

#define false 0
#define true 1

enum OPCODES {
	OP_HALT = 0x00000000,
	OP_RUN = 0x20000000,
	OP_SETMASK = 0x40000000,
	OP_GETINFO = 0x60000000,
	OP_SETRULE = 0x80000000,
	OP_SETEDGE = 0xA0000000,
	OP_LOADCOL = 0xC0000000,
	OP_RST = 0xE0000000
};

unsigned int min(int a, int b) {
	if (a < b) {
		return a;
	}
	return b;
}


unsigned int cam_get_width() {
	return g_grid_width;
}

unsigned int cam_get_height() {
	return g_grid_height;
}

unsigned int cam_get_num_neighbours() {
	return g_num_neighbours;
}




void cam_wait_for_done() {
	int true_value = 1;
	int coproc_done = 0;
	asm (
		"wait_loop:\n\t"
		"mrc 14, 0, %[coproc_done], cr15, cr5\n\t"
		"cmp %[true_value], %[coproc_done]\n\t"
		"bne wait_loop\n\t"
		: [coproc_done]"=r"(coproc_done)
		: [true_value]"r"(true_value)
		:
		);
}

// Internal only, the interface only exposes the
// functions that wrap the various CAM-functions.
void cam_do_operation(unsigned int op) {
	// HACK, FIXME, TODO:
	if (op != 0) {
		cam_do_operation(0);
	}
	cam_wait_for_done();
	asm (
		"mcr 14, 0, %[op], cr15, cr5\n\t"
		:
		: [op]"r"(op)
		:
		);
}

// Internal only, use this to set the operand fields when
// the edge data is needed. convention is, lower bits in lower
// op_num's (cr-indexes).
void cam_set_operand(unsigned int op_num, unsigned int data) {
	cam_wait_for_done();
	switch (op_num) {
		case 0:
			asm ("mcr 14, 0, %[data], cr0, cr5\n\t" : : [data]"r"(data) :);
			break;
		case 1:
			asm ("mcr 14, 0, %[data], cr1, cr5\n\t" : : [data]"r"(data) :);
			break;
		default:
			printf("ERROR: operand_index %d is not implemented currently, update cam_set_operand\n", op_num);
			break;
	}
}

unsigned int cam_read_output(int index) {
	unsigned int output = 0;
	cam_wait_for_done();
	switch (index) {
		case 0:
			asm ("mrc 14, 0, %[output], cr0, cr5\n\t" : [output]"=r"(output) ::);
			break;
		case 1:
			asm ("mrc 14, 0, %[output], cr1, cr5\n\t" : [output]"=r"(output) ::);
			break;
		default:
			printf("ERROR: cam_read_output can not handle output index %d\n", index);
			break;
	}
	return output;
}

void cam_get_information() {
	unsigned int operation = 7;
	operation <<= 28;

	cam_do_operation(operation);
	unsigned int result = cam_read_output(0);
	unsigned int numGrids = (result & 0xFFFF0000) >> 16;
	g_grid_height = result & 0x0000FFFF;
	operation = 6 << 28;

	cam_do_operation(operation);
	result = cam_read_output(0); // OP_GETINFO always writes to the lowest 32 output bits. 
	unsigned int cellsPerSub = (result & 0xFFFF0000) >> 16;
	g_num_neighbours = result & 0x0000FFFF;
	g_grid_width = numGrids * cellsPerSub;
	printf("Result: %08X\n", result);
	printf("Height: %d Width: %d NumGrids: %d cellsPerSub: %d numNeighbours: %d\n", g_grid_height, g_grid_width, numGrids, cellsPerSub, g_num_neighbours);
}

void cam_init() {
	cam_get_information();
	cam_reset_to(0);
}

void cam_reset_to(int value) {
	cam_wait_for_done();
	int operation = 0xe0000000;
	value = (value & 1) << 28;
	operation |= value;
	cam_do_operation(operation);
}

void cam_set_edge(int modeEW, int modeNS, int value) {
	int operation = 0xa0000000;

	operation |= (modeEW & 0x1) << 25;
	operation |= (modeNS & 0x1) << 24;
	// TODO: Un-hardcode
	value &= 0xFF;
	value <<= (23 - g_grid_height + 1);
	operation |= value;
	cam_wait_for_done();
	cam_do_operation(operation);
}

// Remember to set operand before calling
void cam_load_column_internal(unsigned int times) {
	int operation = 0xc0000000;
	times &= 0x1FFFFFFF; // 29 bit.
	operation |= times;
	cam_do_operation(operation);
}

// Only handles lowest 32 bits.
void cam_load_column(unsigned int value, unsigned int times) {
	cam_set_edge(0, 0, 0);
	cam_set_operand(0, value);
	for (int i = 1; i < (g_grid_height / 32 + 1); i++) {
		cam_set_operand(i, 0);
	}
	cam_load_column_internal(times);
}

void cam_load_big_column(const char *column, unsigned int times) {
	// If input string is shorter than the height, we consider that as leading 0s
	// additionally, any non '1' character is a 0.
	// If input string is longer than the height, ignore the last values
	// In sum, this means that the first values are always used, and that we start
	// filling from the lowest index on the grid.
	int length = min(g_grid_height, strlen(column));
	int numOperands = length / 32 + (length % 32 ? 1 : 0);
	unsigned int operand = 0;
	unsigned int operand_num = 0;
	cam_set_edge(0, 0, 0);
	for (int i = length - 1, pos = 0; i >= 0; i--, pos++) {
		int value = (column[i] == '1' ? 1 : 0);
		operand |= (value << (pos));

		if (pos == 31 || i == 0) {
			cam_set_operand(operand_num, operand);
			operand = 0;
			pos = -1; // Is incremented by for loop
			operand_num++;
		}
	}
	// Zero out any untouched operands.
	if (numOperands < (g_grid_height / 32 + 1)) {
		for (int i = numOperands; i < (g_grid_height / 32 + 1); i++) {
			cam_set_operand(i, 0);
		}
	}
	cam_load_column_internal(times);
}

void cam_shift_column(unsigned int times) {
	cam_set_edge(1, 1, 0);
	for (int i = 0; i <= g_grid_height / 32; i++) {
		cam_set_operand(i, 0);
	}
	cam_load_column_internal(times);
}

void cam_run(unsigned int generations) {
	cam_wait_for_done();
	int operation = 0x20000000;
	generations &= 0x1FFFFFFF;
	operation |= generations;
	cam_do_operation(operation);
}

void hex2bin(unsigned int hex, char *result, int size, int prioritize_msb) {
	// If the size is not large enough, then we must choose which end we actually store.
	if (prioritize_msb) {
		unsigned int mask = (1 << (size - 1));
		// Position 0 is MSB.
		for (int i = 0; i < size; i++) {
			result[i] = (hex & mask ? '1':'0');
			mask >>= 1;
		}
	} else {
		// Prioritize the LSB
		unsigned int mask = 1;
		for (int i = size - 1; i >= 0; i--) {
			result[i] = (hex & mask ? '1' : '0');
			mask <<= 1;
		}
	}
}

unsigned int bin2hex(char *hex, int size) {
	int result = 0;
	int multiplier = (1 << (size - 1));
	size = min(size, strlen(hex));
	for (int i = 0; i < size; i++) {
		result += (hex[i] == '1' ? 1 : 0) * multiplier;
		multiplier >>= 1;
	}
	return result;
}

// Side-effect: edge is stuck in wrap mode after print.
void cam_print_grid() {
	cam_wait_for_done();
	cam_set_edge(1, 1, 0);

	char col_str[32];
	memset(col_str, 0, g_grid_height + 1);
	for (int i = 0; i < g_grid_width; i++) {
		printf("Column %2d: ", i);
		for (int j = g_grid_height / 32; j >= 0; j--) {
			unsigned int width = 32;
			// Crop the length of the highest one.
			if (j == g_grid_height / 32) {
				width -= g_grid_height % 32;
			}
			hex2bin(cam_read_output(j), col_str, width, false);
			printf("%s (%08X) ", col_str, cam_read_output(j));
		}
		printf("\n");
		cam_shift_column(1);
	}
}

void cam_print_grid_rotated() {
	cam_wait_for_done();
	size_t grid_size = (g_grid_width + 1) * g_grid_height;
	char *grid = (char*)malloc((g_grid_width + 1) * g_grid_height);
	char *col_str = (char*)malloc(g_grid_height + 1);
	memset(grid, 'X', (g_grid_width + 1) * g_grid_height);
	memset(col_str, 0, g_grid_height + 1);
	
	// Insert 0-terminators.
	for (int i = g_grid_width; i < grid_size; i += (g_grid_width + 1)) {
		grid[i] = 0;
	}
	cam_set_edge(1, 1, 0);
	
	// Reconstruct rows from columns
	for (int col = g_grid_width - 1; col >= 0; col--) {
		// Get a column
		int index = 0;
		for (int i = g_grid_height / 32; i >= 0; i--) {
			unsigned int width = 32;
			// Crop the length of the highest one.
			if (i == g_grid_height / 32) {
				width -= g_grid_height % 32;
			}
			hex2bin(cam_read_output(i), col_str + index, width, false);
			index += width;
		}
		for (int row = 0; row < g_grid_height; row++) {
			grid[(row * (g_grid_width + 1)) + col] = col_str[row];
		}
		cam_shift_column(1);
	}
	for (int row = 0; row < g_grid_height; row++) {
		printf("%s\n", grid + (row * (g_grid_width + 1)));
	}
	free(grid),
	free(col_str);
}

// TODO: This will only work on the upper 32 rows for now.
unsigned int cam_get_row(unsigned int row) {
	unsigned int result = 0;
	unsigned int shift;
	unsigned int width;//, height;
	unsigned int lostColumns = 0;
	if (g_grid_width > 32) {
		width = 32;
		lostColumns = g_grid_width - 32;
	} else {
		width = g_grid_width;
	}

	shift = row;
	cam_set_edge(1, 1, 0);
	for (int i = 0; i < width; i++) {
		unsigned int col = cam_read_output(row / 32);
		int val = (col >> (shift % 32)) & 1;
		val <<= i;
		result |= val;
		cam_shift_column(1);
	}

	for (int i = 0; i < lostColumns; i++) {
		cam_shift_column(1);
	}
	return result;
}

void cam_get_full_row(unsigned int row, unsigned int *dst) {
	unsigned int width = g_grid_width;
	unsigned int shift = row;

	cam_set_edge(1, 1, 0);
	for (int i = 0; i < width; i++) {
		unsigned int col = cam_read_output(row / 32);
		int val = (col >> (shift % 32)) & 1;
		dst[i] = (val ? 1 : 0);
		cam_shift_column(1);
	}
}

void cam_set_mask(unsigned int fullRow, unsigned int fullCol, unsigned int value, unsigned int maskX, unsigned int maskY) {
	unsigned int op = OP_SETMASK & (0xE0000000);
	op |= (fullRow & 1) << 28; 
	op |= (fullCol & 1) << 27;
	op |= (value & 1) << 26;
	op |= (maskX & 0x1ff) << 17;
	op |= (maskY & 0x1ff) << 8;
	cam_do_operation(op);
}

void cam_set_rule_key_value(unsigned int key, unsigned int value) {
	unsigned int operation = OP_SETRULE;
	unsigned int valueMask = 0xff;
	unsigned int valueShift = 21;
	unsigned int keyMask = (1 << (g_num_neighbours - 2)) - 1;
	unsigned int keyShift = valueShift - (g_num_neighbours - 2);
	value &= valueMask;
	value <<= valueShift;
	operation |= value;
	key &= keyMask;
	key <<= keyShift;
	operation |= key;
	cam_do_operation(operation);	
}

void cam_set_wolfram_rule_internal(unsigned int rule, unsigned int north_south) {
	//         W c E
	//       N W c E S
	// NW SW N W c E S NE SE
	unsigned int w_shift = 2;
	unsigned int c_shift = 1;
	unsigned int e_shift = 0;
	if (g_num_neighbours == 8) {
		w_shift += 3;
		c_shift += 3;
		e_shift += 3;
	} else if (g_num_neighbours == 4) {
		w_shift += 1;
		c_shift += 1;
		e_shift += 1;
	} else if (g_num_neighbours == 2) {
		w_shift = 2;
		c_shift = 1;
		e_shift = 0;
	}

	if (north_south) {
		w_shift++;
		e_shift--;
	}

	unsigned char rule_array[64] = {0};
	unsigned int full_rule = 0;
	for (int key = 0; key < (1 << (g_num_neighbours - 2)); key++) {
		full_rule = key << 3;
		unsigned int value = 0;

		for (int i = 0; i < 8; i++) {
			full_rule |= i;
			unsigned int w_bit = (full_rule >> w_shift) & 0x1;
			unsigned int c_bit = (full_rule >> c_shift) & 0x1;
			unsigned int e_bit = (full_rule >> e_shift) & 0x1;

			unsigned int actual_rule = (w_bit << 2) | (c_bit << 1) | e_bit;
			if (rule & (1 << actual_rule)) {
				value |= (1 << i);
			}
		}
		
		rule_array[key] = value;
	}
	cam_set_rule_from_array(rule_array);
}

void cam_set_wolfram_rule_ns(unsigned int rule) {
	cam_set_wolfram_rule_internal(rule, 1);
}

void cam_set_wolfram_rule(unsigned int rule) {
	cam_set_wolfram_rule_internal(rule, 0);
}

void cam_set_rule_from_array(unsigned char *rule_array) {
	for (int i = 0; i < (1 << (g_num_neighbours - 2)); i++) {
		cam_set_rule_key_value(i, rule_array[i]);
	}
}

// Take the northern most row of the grid, and use the current rule
// to extend it so that the rows below contain the following generations.
// This only works for 1D automata
// but it also only makes sense for them.
void cam_create_multi_gen() {
	unsigned int **generations = (unsigned int**)malloc(sizeof(unsigned int*) * g_grid_height);
	for (int y = 0; y < g_grid_height; y++) {
		// Allocate room for this generation.
		generations[y] = (unsigned int*)malloc(sizeof(unsigned int) * g_grid_width);
		for (int x = 0; x < g_grid_width; x++) {
			generations[y][x] = 0;
		}
		cam_get_full_row(g_grid_height - 1, generations[y]);
		
		// Get next generation
		cam_run(1);
	}

	if (g_grid_height <= 32) {
		for (int x = 0; x < g_grid_width; x++) {
			unsigned int column = 0;
			for (int y = 0; y < g_grid_height; y++) {
				// Lower generation values should go in the top bits, and vice verca.
				column |= (generations[g_grid_height - y - 1][x] ? (1 << y) : 0); 
			}
			cam_load_column(column, 1);
		}
	} else { // Use big-load for handling the larger sizes.
		char *column_str = (char*)malloc(sizeof(char) * g_grid_height);
		for (int x = 0; x < g_grid_width; x++) {

			for (int y = 0; y < g_grid_height; y++) {
				// Lower generation values should go in the top bits, and vice verca.
				column_str[g_grid_height - y - 1] = (generations[g_grid_height - y - 1][x] ? '1' : '0'); 
			}
			cam_load_big_column(column_str, 1);
		}
	}
	cam_set_edge(1, 1, 0);
	// Cleanup
	for (int y = 0; y < g_grid_height; y++) {
		free(generations[y]);
	}
	free(generations);
}

void cam_load_single_north_east_1() {
	char *load_array = (char*) malloc(sizeof(char) * (g_grid_height + 1));
	memset(load_array, '0', sizeof(char) * (g_grid_height + 1));
	load_array[g_grid_height] = 0;
	load_array[0] = '1';
	cam_load_big_column(load_array, 1);
	load_array[0] = '0';
	cam_load_big_column(load_array, g_grid_width - 1);
	free(load_array);
}

void cam_load_full_grid(const char *dst) {
	for (int i = 0; i < g_grid_width; i++) {
		cam_load_big_column(dst + g_grid_height * i, 1);
	}
}

void cam_get_full_grid(char *dst) {
	cam_wait_for_done();
	cam_set_edge(1, 1, 0);
	unsigned int pos = 0;
	for (int i = 0; i < g_grid_width; i++) {
		for (int j = g_grid_height / 32; j >= 0; j--) {
			unsigned int width = 32;
			// Crop the length of the highest one.
			if (j == g_grid_height / 32) {
				width -= g_grid_height % 32;
			}
			hex2bin(cam_read_output(j), dst + pos, width, false);
			pos += width;
		}
		cam_shift_column(1);
	}
}

void set_rule_to_game_of_life() {
	int c_shift = 4;
	unsigned char rule_array[64] = {};
	unsigned int value = 0;

	for (int i = 0; i < 512; i++) {
		int live_neighbours = 0;
		int live = (i & (1 << c_shift));
		for (int j = 0; j < 9; j++) {
			// If it's the center cell.
			if (j == c_shift) {
				// Do nothing
			} else if (i & (1 << j)) {
				live_neighbours++;
			}
			if (live_neighbours > 3) {
				break;
			}
		}
		if (live) {
			if (live_neighbours == 2 || live_neighbours == 3) {
				value |= (1 << (i %8));
			}
		} else {
			if (live_neighbours == 2) {
				value |= (1 << (i%8));
			}
		}
		if (i == 0 && i % 8 == 0) {
			rule_array[i / 64] = value;
			value = 0;
		}
	}

	cam_set_rule_from_array(rule_array);
}

char *cam_allocate_grid_array() {
	return (char*)malloc(sizeof(char) * g_grid_width * g_grid_height + 1);
}

