#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cam.h"
#include "random.h"

int test_reset_to_0() {
	unsigned int result = 2;
	// Test: Reset to 0
	printf("Test: Reset to 0\n");
	cam_reset_to(0);	
	unsigned int height = cam_get_height();
	unsigned int width = cam_get_width();
	unsigned int num_outputs = height / 32;
	unsigned int ignore_bits = 32 - (height % 32);
	int failed = 0;
	cam_set_edge(1, 1, 0);
	for (int x = 0; x < width; x++) {
		for (int i = 0; i <= num_outputs; i++) {
			result = cam_read_output(i);
			unsigned int expected = 0x00000000;
			if (ignore_bits > 1 && i == num_outputs) {
				expected >>= ignore_bits;
			}
			if (result != expected) {
				printf("Test fail at output(%d): Expected %08X, Got %08X\n", i, expected, result);
				failed = 1;
			}
		}
		cam_load_column(0xFFFFFFFF, 1);
	}
	printf("Result: %08X\n", result);
	// Result should be 0	
	if (failed) {
		printf("Test failed\n");
	} else {
		printf("Test suceeded\n");
	}
	printf("End of test: Reset to 0\n\n");
	return failed;
}

int test_reset_to_1() {
	unsigned int result = 2;
	// Test: Reset to 1.
	printf("Test: Reset to 1\n"); 
	cam_reset_to(1);
	unsigned int height = cam_get_height();
	unsigned int width = cam_get_width();
	unsigned int num_outputs = height / 32;
	unsigned int ignore_bits = 32 - (height % 32);
	int failed = 0;
	cam_set_edge(1, 1, 0);
	for (int x = 0; x < width; x++) {
		for (int i = 0; i <= num_outputs; i++) {
			result = cam_read_output(i);
			unsigned int expected = 0xFFFFFFFF;
			if (ignore_bits > 1 && i == num_outputs) {
				expected >>= ignore_bits;
			}
			if (result != expected) {
				printf("Test fail at col(%d) output(%d): Expected %08X, Got %08X\n", x, i, expected, result);
				failed = 1;
			}
		}
		cam_load_column(0, 1);
	}
	printf("Result: %08X\n", result);
	// Result should be 255
	if (failed) {
		printf("Test failed\n");
	} else {
		printf("Test suceeded\n");
	}
	printf("End of test: Reset to 1\n\n");
	return failed;
}

void test_random_rule_30(int num_numbers) {
	printf("Starting test: Rule 30 Pseudo Random\n");
	cam_reset_to(0);
	cam_set_edge(0, 0, 0);
	cam_set_wolfram_rule(30);
	cam_load_column(bin2hex("10000000", 8), 1);
	cam_load_column(bin2hex("00000000", 8), 7);
	cam_set_edge(1, 1, 0);

	cam_create_multi_gen();
	cam_print_grid_rotated();
	for (int i = 0; i < num_numbers; i++) {
		printf ("%d%s", cam_read_output(0), (i % 20 != 0 ? ", " : "\n"));
		cam_run(8);
	}
	printf("End of test rule 30 Pseudo Random\n");
}

void test_wolfram_rules_multigen() {
	cam_reset_to(0); // Make sure we have a good initial state, with no state masks etc.
	int height = cam_get_height();
	int width = cam_get_width();
	printf("Test: Wolfram rules 0-255, %d generations, init value: single 1\n", height);
	char *init_row = (char*)malloc(sizeof(char) * (height + 1));
	memset(init_row, '0', height + 1);
	init_row[0] = '1';
	init_row[height] = '\0';
	for (int i = 30; i < 31; i++) {
		cam_set_edge(0, 0, 0); // Prepare for load
		cam_load_big_column(init_row, 1);
		cam_load_column(0, width - 1);
		cam_set_edge(1, 1, 0);
		cam_set_wolfram_rule(i);
		cam_create_multi_gen();
		printf("Rule %d: After multigen (implicitly %d generations)\n", i, height);
		cam_print_grid_rotated();
		cam_run(height);
		printf("Rule %d: After additional %d generations:\n", i, height);
		cam_print_grid_rotated();
	}
	free(init_row);
}

void test_wolfram_rule(unsigned int rule, unsigned int generations, int do_reset) {
	printf("Test: Wolfram rule %d, %d generations, init value: single 1 reset first: %s\n", rule, generations, (do_reset ? "true" : "false"));
	if (rule > 255) {
		printf("ERROR: %d is not a valid Wolfram rule number\n", rule);
		return;
	}
	if (do_reset) {
		cam_reset_to(0);
	}
//	int height = cam_get_height();
	int width = cam_get_width();
	cam_set_edge(0, 0, 0);
	cam_load_column(bin2hex("11111111", 8), 1);
	cam_load_column(0, width - 1);
	cam_set_wolfram_rule(rule);
	cam_set_edge(1, 1, 0);
	cam_run(generations);
	printf(" Rule: %d - After %d generations\n", rule, generations);
	cam_print_grid_rotated();
	printf("End of Wolfram-test rule %d, %d generations\n", rule, generations);
}

void test_wolfram_rules(int generations) {
	printf("Test: Wolfram rules 0-255, %d generations, init value: single 1\n", generations);
	cam_reset_to(0);
	for (int i = 0; i < 255; i++) {
		test_wolfram_rule(i, generations, 0);
	}
	printf("End of full Wolfram-test %d generations\n", generations);
}

void load_gospel_glider() {
	cam_load_big_column("00000000000000000000000000000000000000", 1);
	cam_load_big_column("00000000000000000000000001000000000000", 1);
	cam_load_big_column("00000000000000000000000101000000000000", 1);
	cam_load_big_column("00000000000001100000011000000000000000", 1);
	cam_load_big_column("00000000000010001000011000000000000110", 1);
	cam_load_big_column("01100000000100000100011000000000000110", 1);
	cam_load_big_column("01100000000100010110000101000000000000", 1);
	cam_load_big_column("00000000000100000100000001000000000000", 1);
	cam_load_big_column("00000000000010001000000000000000000000", 1);
	cam_load_big_column("00000000000001100000000000000000000000", 1);
	cam_load_big_column("00000000000000000000000000000000000000", 4);
}

void test_gospel_glider_print_final_only(unsigned int generations) {
	printf("Test: Gospel glider print final only, %d generations\n", generations);
	cam_reset_to(0);
	set_rule_to_game_of_life();
	load_gospel_glider();
	cam_set_edge(0, 0, 0);
	cam_run(generations);
	cam_print_grid_rotated();
	printf("End of test: Gospel glider, print final only generations\n");
}

void test_gospel_glider_all_generations(unsigned int generations) {
	printf("Test: Gospel glider print all generations, %d generations\n", generations);
	cam_reset_to(0);
	set_rule_to_game_of_life();
	load_gospel_glider();
	cam_set_edge(0, 0, 0);
	for (int i = 0; i < generations; i++) {
		cam_run(1);
		cam_print_grid_rotated();
	}
	printf("End of test: Gospel glider, print all generations\n");
}

// Stand-alone density classification with
void test_density_classification_rule_184_232_ns() {
	printf("Start of test: Density classification using rule 184 and rule 232 in North/South-direction, and input created by Rule 30\n");
	// http://arxiv.org/pdf/comp-gas/9703001v1.pdf
	int width = cam_get_width();
	int height = cam_get_height();

	// Use a random-solution to generate the necessary amount of random numbers.
	cam_reset_to(0);
	cam_load_single_north_east_1();
	cam_set_wolfram_rule(30);
	cam_create_multi_gen();
	cam_set_edge(1,1,0);

	int random_bits_needed = height;
	char **random_columns = (char**)malloc(sizeof(char*) * width);
	int *count = (int*)malloc(sizeof(int) * width);
	memset(count, 0, sizeof(int) * width);

	for (int x = 0; x < width; x++) {
		random_columns[x] = (char*)malloc(sizeof(char) * (random_bits_needed + 1));
		memset(random_columns[x], 0, sizeof(char) * (random_bits_needed + 1));

		int random_bits_per_iteration = (height > 32 ? 32 : height);
		for (int i = 0; i < random_bits_needed; i += random_bits_per_iteration) {
			cam_run(height);
			unsigned int random = cam_read_output(0);
			for (int j = 0; j < random_bits_per_iteration; j++) {
				if (i + j >= random_bits_needed) {
					break;
				}
				unsigned int bit = ((random >> j) & 1);
				count[x] += bit;
				random_columns[x][i + j] = (bit ? '1' : '0');
			}
		}
	}
	cam_set_edge(0, 0, 0);
	for (int x = 0; x < width; x++) {
		cam_load_big_column(random_columns[x], 1);
	}

	// Final formula from II, and repeated as Propositon 4.
	int rule_184_generations = ((height - 2) / 2);
	int rule_232_generations = ((height - 1) / 1);

	cam_set_edge(1,1,0);
	cam_set_wolfram_rule_ns(184);
	cam_run(rule_184_generations);
	cam_set_wolfram_rule_ns(232);
	cam_run(rule_232_generations);

	// Verification:
	printf("Verification, now the result for each row (or an error) will be listed with 0 = most 0, 1 = most 0, X = equal\n");
	cam_set_edge(1, 1, 0);
	for (int x = 0; x < width; x++) {
		int should_be = 0;
		if (count[x] > height/2) {
			should_be = 1;
		} else if (count[x] == height/2 && height % 2 == 0) {
			should_be = 2;
		} else {
			should_be = 0;
		}
		free(random_columns[x]);
		int failure = 0;
		for (int i = 0; i < (height/32 + 1); i++) {
			unsigned int output = cam_read_output(i);
			unsigned int ignore_bits = 0;
			if (i == (height/32) && height % 32) {
				ignore_bits = 32 - (height % 32);
			}
			if (output != 0 && should_be == 0) {
				printf("ERROR: This column should be 0, but was something else %08X\n", output);
				failure = 1;
				break;
			} else if (output != (0xFFFFFFFF >> ignore_bits) && should_be == 1) {
				printf("ERROR: This column should be 1, but was something else %08X\n", output);
				failure = 1;
				break;
			} else if (should_be == 2) {
				int bit = output & 1;
				for (int j = 1; j < (32 - ignore_bits); j++) {
					if (((output >> j) & 1) != !bit) {
						printf("ERROR: This column should be alternating, but was something else %08X\n", output);
						failure = 1;
						break;
					}
					bit = (output >> j) & 1;
				}
			}
		}
		if (!failure) {
			if (should_be == 0) {
				printf("0");
			} else if (should_be == 1) {
				printf("1");
			} else {
				printf("X");
			}
		}
		cam_load_column(0, 1); // Shift
	}
	printf("\n");
	free(random_columns);

	free(count);
	printf("End of test: Randomized column-wise density classification\n");
}

int test_density_classification_rule_184_232_ns_r30rand() {
	printf("Start of test: Density classification using rule 184 and rule 232 in North/South-direction, and input created by Rule 30 (random.c)\n");
	// http://arxiv.org/pdf/comp-gas/9703001v1.pdf
	int width = cam_get_width();
	int height = cam_get_height();

	// Use a random-solution to generate the necessary amount of random numbers.

	int random_bits_needed = height;
	char **random_columns = (char**)malloc(sizeof(char*) * width);
	int *count = (int*)malloc(sizeof(int) * width);
	memset(count, 0, sizeof(int) * width);

	for (int x = 0; x < width; x++) {
		random_columns[x] = (char*)malloc(sizeof(char) * (random_bits_needed + 1));
		memset(random_columns[x], 0, sizeof(char) * (random_bits_needed + 1));

		int random_bits_per_iteration = (height > 32 ? 32 : height);
		for (int i = 0; i < random_bits_needed; i += random_bits_per_iteration) {
			cam_run(height);
			unsigned int random = rng_get_uint();
			for (int j = 0; j < random_bits_per_iteration; j++) {
				if (i + j >= random_bits_needed) {
					break;
				}
				unsigned int bit = ((random >> j) & 1);
				count[x] += bit;
				random_columns[x][i + j] = (bit ? '1' : '0');
			}
		}
	}
	cam_set_edge(0, 0, 0);
	for (int x = 0; x < width; x++) {
		cam_load_big_column(random_columns[x], 1);
	}
#if 0 // For debugging
	printf("Initial state:\n");
	cam_print_grid_rotated();
	printf("\n");
#endif
	// Final formula from II, and repeated as Propositon 4.
	int rule_184_generations = ((height - 2) / 2);
	int rule_232_generations = ((height - 1) / 1);

	cam_set_edge(1,1,0);
	cam_set_wolfram_rule_ns(184);
	cam_run(rule_184_generations);
	cam_set_wolfram_rule_ns(232);
	cam_run(rule_232_generations);
#if 0 // For debugging
	printf("Final result\n");
	cam_print_grid_rotated();
	printf("\n");
#endif
	// Verification:
	printf("Verification, now the result for each row (or an error) will be listed with 0 = most 0, 1 = most 0, X = equal\n");
	cam_set_edge(1, 1, 0);
	int success = 1;
	for (int x = 0; x < width; x++) {
		int should_be = 0;
		if (count[x] > height/2) {
			should_be = 1;
		} else if (count[x] == height/2 && height % 2 == 0) {
			should_be = 2;
		} else {
			should_be = 0;
		}
		free(random_columns[x]);
		int failure = 0;
		for (int i = 0; i < (height/32 + 1); i++) {
			unsigned int output = cam_read_output(i);
			unsigned int ignore_bits = 0;
			if (i == (height/32) && height % 32) {
				ignore_bits = 32 - (height % 32);
			}
			if (output != 0 && should_be == 0) {
				printf("ERROR: This column should be 0, but was something else %08X\n", output);
				failure = 1;
				break;
			} else if (output != (0xFFFFFFFF >> ignore_bits) && should_be == 1) {
				printf("ERROR: This column should be 1, but was something else %08X\n", output);
				failure = 1;
				break;
			} else if (should_be == 2) {
				int bit = output & 1;
				for (int j = 1; j < (32 - ignore_bits); j++) {
					if (((output >> j) & 1) != !bit) {
						printf("ERROR: This column should be alternating, but was something else %08X\n", output);
						failure = 1;
						break;
					}
					bit = (output >> j) & 1;
				}
			}
		}
		if (!failure) {
			if (should_be == 0) {
				printf("0");
			} else if (should_be == 1) {
				printf("1");
			} else {
				printf("X");
			}
		} else {
			success = 0;
		}
		cam_load_column(0, 1); // Shift
	}
	printf("\n");
	free(random_columns);

	free(count);
	printf("End of test: Randomized column-wise density classification, result %s\n", (success ? "SUCCESS" : "FAILURE"));
	return success;
}

void test_density_184_232_n_times(unsigned int times) {
	int successes = 0;
	for (int i = 0; i < times; i++) {
		printf("Test iteration %d, successes: %d\n", i, successes);
		successes += test_density_classification_rule_184_232_ns_r30rand();
	}
	printf("Final result: %d/%d tests SUCCEEDED\n", successes, times);
}

void load_gardner_examples() {
	char examples[] = {
	"0000000000000000000000000000000000000000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000001000000001000000000000000000000000000\0"
	"0000010000000001000000000000000000100000000\0"
	"0000001000000001000000001100000001100000000\0"
	"0000000000000000000000011000000000100000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000000000000010000000000000000000000000000\0"
	"0000000000000011000000000000000000000000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000000000000000000000001000000000000000000\0"
	"0000001000000100000000001000000000100000000\0"
	"0000110000000010000000001000000000100000000\0"
	"0000000000000001000000001000000001100000000\0"
	"0000000000000000000000000000000000000000000\0"
	"0000000000000000000000000000000000000000000\0"
	};
	int col_size = 44; // 43 elements + \0
	int num_cols = 18;
	for (int i = 0; i < num_cols; i++) {
		cam_load_big_column(examples + (col_size*i), 1);
	}
}

void load_gosper_glider_gun() {
	char gosper[] = {
	"00000000000000000000000000000000000000\0"
	"00000000000000000000000001000000000000\0"
	"00000000000000000000000101000000000000\0"
	"00000000000001100000011000000000000110\0"
	"00000000000010001000011000000000000110\0"
	"01100000000100000100011000000000000000\0"
	"01100000000100010110000101000000000000\0"
	"00000000000100000100000001000000000000\0"
	"00000000000010001000000000000000000000\0"
	"00000000000001100000000000000000000000\0"
	"00000000000000000000000000000000000000\0"
	"00000000000000000000000000000000000000\0"};
	int col_size = 39; // 38 elements + \0
	int num_cols = 12;
	for (int i = 0; i < num_cols; i++) {
		cam_load_big_column(gosper + (col_size*i), 1);
	}
}
