#include <iostream>
#include <vector>
#include "system.h"
#include "rules.h"
#include "utils.h"
#include "rulesets.h"
#include "gameoflife.h"
#include "opcodes.h"

using namespace std;

#define WIDTH 5*10
#define HEIGHT 1

void runGenerations(System &system, int numIterations, bool printEveryGen = true) {
	for (int i = 0; i < numIterations; i++) {
		if (printEveryGen) {
			system.printSystem();
		}
		system.updateSystem();
	}
	system.printSystem();
	std::cout << endl;
}

bool parseUIntParameter(std::string param, unsigned int &result) {
	if (param.size() == 0 || !isdigit(param.at(0))) {
		return false;
	}
	result = atoi(param.c_str());
	return true;
}

// TODO: Add support for directly rereading the generated assembly?
void printHelp() {
	cout << "Usage: ./simulator [args]" << endl
		<< "-w[integer]  - width (optional, defaults to: " << WIDTH << ")" << endl
		<< "-h[integer]  - height (optional, defaults to: " << HEIGHT << ")" << endl
		<< "-l[01]       - print as LST if 1, (optional, defaults to 1)" << endl
		<< "-g[integer]  - # generations to run after loading" << endl
		<< "-d           - don't run any generations (usefull for VHDL-output)" << endl
		<< "-i[filename] - filename to use for grid-initialization, (optional, defaults to \"input\")" << endl
		<< "-n[3,5,9]    - # neighbour-connections per cell (3 is 1D, 5/9 is 2D connected)" << endl
		<< "-p           - print verbosity 0 = none, 1 = only generations, 2 = include load, defaults to 2 " << endl
		<< "-v           - output VHDL testbench-code" << endl
		<< "-a           - output as Pseudo-Assembly" << endl
		<< "-f[filename] - run Pseudo-Assembly-file (gets priority), and output assembly" << endl
		<< "-e[01][01]   - enable edge-wrapping, first digit is East/West, second is (optionally) North/South, defaults to 00" << endl
        << "-rw[integer] - define rule by Wolfram-number [0-255] (A rule definition is required)" << endl
		<< "-rg          - define rule to be Conway's Game of Life (A rule definition is required)" << endl
		<< "-u           - print only unique generations (works only for 1D, in LST-mode)" << endl;
}

int main(int argc, char **argv) {
	bool printLst = true;
	unsigned int width = WIDTH;
	unsigned int height = HEIGHT;
	unsigned int numNeighbours = 9;
	std::string inputFileName = "input";
	std::string pseudoAssemblyFileName = "";
	int numGenerations = 500;
	bool outputVHDL = false;
	bool outputAssembly = false;
	int wolframRule = -1;
	int testRule = -1;
	bool useGameOfLife = false;
	bool runGenerations = true;
	bool wrapNS = false;
	bool wrapEW = false;
	int printVerbosity = 2;
	bool printUnique = false;

	for (int i = 0; i < argc; i++) {
		std::string arg(argv[i]);
		if (arg.at(0) == '-' && arg.size() > 1) {
			char param = tolower(arg.at(1));
			arg = arg.substr(2, string::npos);
			unsigned int result = 0;
			switch (param) {
				case 'w':
					if (parseUIntParameter(arg, result)) {
						width = result;
					}
					break;
				case 'h':
					if (parseUIntParameter(arg, result)) {
						height = result;
					}
					break;
				case 'l':
					if (parseUIntParameter(arg, result)) {
						printLst = (bool)result;
					}
					break;
				case 'e':
					if (arg.size() > 0) {
						if (parseUIntParameter(arg, result)) {
							wrapEW = result;	
						}
					}
					if (arg.size() > 1) {
						arg = arg.substr(1, string::npos);
						if (parseUIntParameter(arg, result)) {
							wrapNS = result;
						}
					}
					break;
				case 'g':
					if (parseUIntParameter(arg, result)) {
						numGenerations = result;
					}
					break;
				case 'd':
					runGenerations = false;
					break;
				case 'i':
					inputFileName = arg;
					break;
				case 'f':
					pseudoAssemblyFileName = arg;
					break;
				case 'n':
					if (parseUIntParameter(arg, result)) {
						numNeighbours = result;
					}
					break;
				case 'p':
					if (parseUIntParameter(arg, result)) {
						printVerbosity = result;
					}
					break;
				case 'u':
					printUnique = true;
					break;
				case 'v':
					outputVHDL = true;
					break;
				case 'a':
					outputAssembly = true;
					break;
				case 'r':
					if (arg.size() != 0) {
						if (arg.at(0) == 'w' && arg.size() > 1) {
							arg = arg.substr(1, string::npos);
							if (parseUIntParameter(arg, result)) {
								if (result <= 255) {
									wolframRule = result;
									testRule = -1;
								}
							}
						} else if (arg.at(0) == 't' && arg.size() > 1) {
							arg = arg.substr(1, string::npos);
							if (parseUIntParameter(arg, result)) {
								if (result <= 255) {
									wolframRule = -1;
									testRule = result;
								}
							}
						} else if (arg.at(0) == 'g') {
							useGameOfLife = true;
						} else {
							cout << "Unhandled way to define rule-set" << endl;
							return 0;
						}
					} else {
						cout << "Unhandled way to define rule-set" << endl;
						return 0;
					}
					break;
				default:
					printHelp();
					return 0;
					break;
			}
		}
	}

	// Special case when running pseudo-assembly
	// as the pseudo-assembly code creates it's own system-instance
	// and we only want to allow a few settings to affect this:
	if (pseudoAssemblyFileName != "") {
		runProgram(pseudoAssemblyFileName);
		return 0;
	}

	// Verify that the settings are usable:
	if (numNeighbours != 3 && numNeighbours != 5 && numNeighbours != 9) {
		cout << "Num Neighbours " << numNeighbours << " not handled by current implementation" << endl;
		return 0;
	}

	System testSystem(width, height, numNeighbours);
	testSystem.setPrintMode((printLst ? PRINTMODE_LST : PRINTMODE_TERM));
	testSystem.setPrintContinuously(false);
	testSystem.setWrapEW(wrapEW);
	testSystem.setWrapNS(wrapNS);

	Rules *rules = nullptr;
	if (useGameOfLife) {
		rules = generateRuleSet(RULE_GAMEOFLIFE);
	} else if (wolframRule != -1) {
		rules = generateWolframRule(wolframRule);
	} else if (testRule != -1) {
		rules = generateTestRuleSet(testRule, numNeighbours);
	} else {
		cout << "No ruleset defined" << endl;
		printHelp();
		return 0;
	}
	rules->extendToNeighbours(numNeighbours);

	testSystem.resetSystemTo(0);
	testSystem.setRuleSet(rules);
	testSystem.loadFileIntoSystem(inputFileName);
	testSystem.setPrintUniqueOnly(printUnique);

	if (printVerbosity < 2) {
		testSystem.clearHistory(); // Remove load-lines from log.
	}

	if (runGenerations) {
		for (int i = 0; i < numGenerations; i++) {
			testSystem.updateSystem();
		}
	}

	if (printVerbosity > 0) {
		testSystem.printHistory();
	}
	if (outputVHDL) {
		testSystem.outputStateAsVHDLTB(true);
		cout << "-- Configure edge" << endl;
		cout << "\tOPCODE(31 downto 29) <= (OP_SETEDGE);" << endl;
		cout << "\tOPCODE(25) <= '" << (wrapEW ? 1 : 0) << "';" << endl; // Wrap E/W
		cout << "\tOPCODE(24) <= '" << (wrapNS ? 1 : 0) << "';" << endl; // Wrap N/S
		cout << "\tWAIT UNTIL DONE = '1';" << endl;

		cout << "-- Start running" << endl;
		cout << "\tOPCODE(31 downto 29) <= (OP_RUN);" << endl;
		cout << "\tOPCODE(28 downto 13) <= conv_std_logic_vector(" << numGenerations << ",16);" << endl;
		cout << "\tWAIT UNTIL DONE = '1';" << endl;
	}
	
	if (outputAssembly) {
		testSystem.outputStateAsPseudoAssembly(true);
		cout << "/* Configure edge */" << endl;
		// TODO: Support actually setting the edge from here?
		cout << "OP_SETEDGE(" << (wrapEW ? 1 : 0) << "," << (wrapNS ? 1 : 0) << ",0);" << endl;

		cout << "/* Start running */" << endl;
		cout << "OP_RUN(" << numGenerations << ");" << endl;
	}

	return 0;
}
