#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <iomanip>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <stdexcept>
#include "utils.h"
#include "opcodes.h"

using namespace std;

// This is a sample program for testing in XCode, without having to set up paths correctly.
// A lot of the macros depend on having set height/neighbours/width correctly
string g_program = "#height(8);#neighbours(2);#width(8); OP_RST(1);OP_SETEDGE(0,0,0);OP_SETRULE(00011110);OP_LOADCOL(11011100,1);OP_LOADCOL(01100100,1);OP_LOADCOL(00110100,1);OP_LOADCOL(00011101,1);OP_LOADCOL(00000110,1);OP_LOADCOL(00010010,1);OP_LOADCOL(00110011,1);OP_LOADCOL(01010101,1);OP_SETEDGE(1,1,0);"
	"OP_RUN(1);OP_VERIFY_ALL();"
	"OP_RUN(20);OP_VERIFY_ALL();"
	"OP_RUN(30);OP_VERIFY_ALL();"
	"OP_RUN(50);OP_VERIFY_ALL();";

string generateRunBlock() {
	stringstream assembly;
	assembly << "b endJmpFunc" << endl;
	// Param: (r0: opcode-string, containing the num-iterations)
	// Destroys: r10, r1
	// Result: (r0: Resulting output)
	assembly << "runCellFuncLoop:" << endl;
	assembly << "ldr r10, =0x00000001" << endl;
	assembly << "mrc 14, 0, r8, cr15, cr5" << endl;
	assembly << "cmp r10, r8" << endl;
	assembly << "bne runCellFuncLoop" << endl;
	assembly << "mov pc, lr" << endl;
	assembly << "runCellFunc:" << endl;
	assembly << "mov r9, lr" << endl;
	assembly << "mcr 14, 0, r0, cr15, cr5" << endl;
	assembly << "bl runCellFuncLoop" << endl;
	assembly << "mrc 14, 0, r0, cr0, cr5" << endl;
	assembly << "ldr r10, =0x00000000" << endl;
	assembly << "mcr 14, 0, r10, cr15, cr5 @ HALT" << endl; // FIXME
	assembly << "bl runCellFuncLoop" << endl;
	assembly << "mov lr, r9" << endl;
	assembly << "mov pc, lr" << endl;
	assembly << "endJmpFunc:" << endl;
	assembly << "b endProg" << endl;
	assembly << "failTest:" << endl; // TODO
	assembly << "ldr r0, =0xDEADDEAD" << endl;
	assembly << "b failTest" << endl;
	assembly << "endProg:" << endl;
	assembly << "b endProg" << endl;
	return assembly.str();
}

void Operation::assertNumArgs(int num) const {
	if (_tokens.size() - 1 < num) {
		stringstream ss;
		ss << "ERROR: mismatch in number of arguments for operation";
		ss << _tokens.at(0);
		throw logic_error(ss.str());
	}
}
string Operation::getArg(int num) const {
	if (num >= 0 && num < _tokens.size() - 1) {
		return _tokens.at(num + 1);
	} else {
		stringstream ss;
		ss << "ERROR: mismatch in number of arguments for operation";
		ss << _tokens.at(0);
		throw logic_error(ss.str());
	}
}
int Operation::getIntArg(int num) const {
	return atoi(getArg(num).c_str());
}

OPCODES Operation::getOpcode() const {
	string op = _tokens.at(0);
	OPCODES opcode;
	if (op == "OP_HALT") {
		opcode = OP_HALT;
	} else if (op == "OP_RUN") {
		opcode = OP_RUN;
	} else if (op == "OP_SETRULE") {
		opcode = OP_SETRULE;
	} else if (op == "OP_SETEDGE") {
		opcode = OP_SETEDGE;
	} else if (op == "OP_LOADCOL") {
		opcode = OP_LOADCOL;
	} else if (op == "OP_RST") {
		opcode = OP_RST;
	} else if (op == "OP_ASSERT") {
		opcode = OP_ASSERT;
	} else if (op == "OP_PRINT") {
		opcode = OP_PRINT;
	} else if (op == "OP_VERIFY") {
		opcode = OP_VERIFY;
	} else if (op == "OP_VERIFY_ALL") {
		opcode = OP_VERIFY_ALL;
	} else if (op == "OP_MAKEFUNC") {
		opcode = OP_MAKEFUNC;
	} else if (op == "#height") {
		opcode = OP_HEIGHT;
	} else if (op == "#width") {
		opcode = OP_WIDTH;
	} else if (op == "#neighbours") {
		opcode = OP_NEIGHBOURS;
	} else {
		stringstream ss;
		throw logic_error("ERROR: Unknown opcode");
	}
	return opcode;
}


Program::Program(const string& str) : _height(-1), _width(-1), _numNeighbours(-1) {
	string program = trimWhiteSpace(str);
	program = trimComments(program);
	vector<string> tokens = tokenizeString(program, ";");
	try {
		for (auto it = tokens.begin(); it != tokens.end(); ++it) {
			Operation op(*it);
			if (op.isMacro()) {
				handleMacro(op);
			} else {
				// Verify that the operation is valid:
				// getOpcode currently does this validation:
				op.getOpcode();
				_operations.push_back(op);
			}
		}
	} catch (exception& e) {
		cerr << e.what();
		exit(-1);
	}
}

void Program::handleMacro(const Operation &op) {
	if (op.getOpcode() == OP_HEIGHT) {
		op.assertNumArgs(1);
		_height = op.getIntArg(0);
	} else if (op.getOpcode() == OP_WIDTH) {
		op.assertNumArgs(1);
		_width = op.getIntArg(0);
	} else if (op.getOpcode() == OP_NEIGHBOURS) {
		op.assertNumArgs(1);
		_numNeighbours = op.getIntArg(0);
	}
}

Operation::Operation(const std::string &instruction) {
	_tokens = tokenizeString(instruction, "(,)");
	if (_tokens.size() == 0) {
		throw logic_error("ERROR: Operation empty");
	}
}

string Operation::getAssemblyForOperation(Program *prog, bool useRunBlock) {
	OPCODES opcode = getOpcode();
	unsigned int iResult = opcode << 29;

	switch (opcode) {
		case OP_HALT:
			// No args
			break;
		case OP_RUN:
			assertNumArgs(1);
			iResult |= getIntArg(0) << 13;
			// [numIterations][28 downto 13]
			break;
		case OP_SETRULE:
			if (prog->getNumNeighbours() == -1) {
				throw logic_error("ERROR while parsing OP_SETRULE, numNeighbours undefined");
			}
			if (prog->getNumNeighbours() > 2) { // untested
				assert(0); // test me
				assertNumArgs(2);
				int ruleKey = getIntArg(0);
				if (ruleKey > ((1 << (prog->getNumNeighbours() + 1)) - 1)) {
					throw logic_error("ERROR while parsing OP_SETRULE");
				}
				iResult |= ruleKey << 28;
			}
			{
				assertNumArgs(1);
				string val = getArg(0);
				int ruleVal;
				if (val.length() == 8) {
					// Attempt to parse it as binary.
					ruleVal = bin2dec(val);
				} else {
					ruleVal = getIntArg(0);
				}
				if (ruleVal > 0xFF) {
					throw logic_error("ERROR while parsing OP_SETRULE");
				}
				if (prog->getNumNeighbours() > 2) {
					iResult |= ruleVal << (28 - prog->getNumNeighbours() - 3);
				} else {
					iResult |= ruleVal << (21);
				}
			}
			// [ruleKey] [28 downto 28 - numNeighbours - 3] [ruleVal][8 next bits]
			break;
		case OP_SETEDGE:
			// [modeEW][25] [modeNS][24] [value][23 downto height - 1]
			assertNumArgs(3);
			(getIntArg(0) ? iResult |= (1 << 25): 0);
			(getIntArg(1) ? iResult |= (1 << 24): 0);
			// TODO: Implement setting the edge value through OP_SETEDGE
			if (getIntArg(2) != 0) {
				assert(0); // TODO: Implement;
			}
			break;
		case OP_LOADCOL: {
			assertNumArgs(2);
			// Support single 0 as fully 0
			string value = getArg(0);
			int intValue = 0;
			if (value != "0") {
				intValue = bin2dec(value.c_str());
			}
			iResult |= intValue << (28 - prog->getHeight() + 1);
			iResult |= atoi(getArg(1).c_str()) << (28 - prog->getHeight() - 8 + 1);
 
			// [value][28 downto height - 1] [times][28 - height downto height - 8 + 1]
			break;
		}
		case OP_RST:
			assertNumArgs(1);
			((getIntArg(0)) ? iResult |= (1 << 28): 0);
			// value[28]
			break;
		case OP_ASSERT: {
			// This depends on the prior instruction having placed the result in r0,
			// otherwise we could have done: ss << "mrc 14, 0, r0, cr0, cr5" << endl;
			assertNumArgs(1);
			stringstream ss;
			ss << "ldr r3, =0x" << std::hex << std::setw(8) << std::setfill('0')
			   << getIntArg(0) << std::dec
			   << "@ OP_ASSERT(" << getIntArg(0) << ")" << endl;
			ss << "mov r2, r0" << endl;
			ss << "cmp r2, r3" << endl;
			ss << "bne failTest" << endl;
			return ss.str();
		}
		// TODO: Make this just insert a proper OP_ASSERT?
		case OP_VERIFY_ALL: {
			Operation edge("OP_SETEDGE(1,1,0)");
			Operation shift("OP_LOADCOL(0,1)");
			stringstream ss;
			ss << edge.getAssemblyForOperation(prog);
			for (int i = 0; i < prog->getWidth(); i++) {
				std::string value = popResult();
				ss << "ldr r3, =0x" << std::hex << std::setw(8) << std::setfill('0')
				   << bin2dec(value) << std::dec
				   << "@ OP_VERIFY(" << bin2dec(value) << ")" << endl;
				ss << "mov r2, r0" << endl;
				ss << "cmp r2, r3" << endl;
				ss << "bne failTest" << endl;
				ss << shift.getAssemblyForOperation(prog);
			}
			return ss.str();
		}
		case OP_VERIFY: {
			stringstream ss;
			std::string value = popResult();
			ss << "ldr r3, =0x" << std::hex << std::setw(8) << std::setfill('0')
			   << bin2dec(value) << std::dec
			   << "@ OP_VERIFY(" << bin2dec(value) << ")" << endl;
			ss << "mov r2, r0" << endl;
			ss << "cmp r2, r3" << endl;
			ss << "bne failTest" << endl;
			return ss.str();
		}
		// Assembly only:
		case OP_MAKEFUNC:
		// TODO: Rewrite to a configuration?
			return generateRunBlock();
			break;
		// Simulator only:
		case OP_PRINT:
			return "";
			break;
		default:
			throw logic_error("ERROR: Unknown OPCODE");
	};

	stringstream ss;
	ss << "ldr r0, =";
	ss << "0x" << std::hex  << std::setw(8) << std::setfill('0') << iResult;
	ss << " @ " << _tokens.at(0) << endl;
	if (useRunBlock) {
		ss << "bl runCellFunc" << endl;
	} else {
		ss << "mcr 14, 0, r0, cr15, cr5" << endl;
	}
	return ss.str();
}

bool Operation::isMacro() const {
	if (_tokens.at(0).at(0) == '#') {
		return true;
	}
	return false;
}

string Program::convertToArmAssembly(bool doGenerateRunBlock) {
	stringstream assembly;
	try {
		for (auto it = _operations.begin(); it != _operations.end(); ++it) {
			string result = it->getAssemblyForOperation(this);
			// Simulator-only instructions return "".
			if (result != "") {
				assembly << result;
			}
		}
	} catch (exception& e) {
		cerr << e.what();
		exit(-1);
	}
	return assembly.str();
}

void Operation::pushResult(const std::string &str) {
	_results.push_back(str);
}

std::string Operation::popResult() {
	if (_results.empty()) {
		throw logic_error("ERROR: Result stack underflow, did you forget to run the program in simulator first?");
	}
	string result = _results.front();
	_results.pop_front();
	return result;
}

string loadFile(string filename);

void runProgram(string filename) {
	string program = loadFile(filename);
	Program p(program);
	p.runInSimulator();
	cout << p.convertToArmAssembly(true) << endl;
}

// Optional main-function for separate compilation instead of merging into
// the increasingly complex "simulator".
int main2(int argc, char **argv) {
	string program;
	if (argc > 1) {
		string filename = argv[1];
		program = loadFile(filename);
	} else {
		program = g_program;
	}
	Program p(program);
	p.runInSimulator();
	cout << p.convertToArmAssembly(true) << endl;
	return 0;
}
