#include <iostream>
#include <cmath>
#include <cassert>
#include "rules.h"
#include "utils.h"

int loadModePosition;

using namespace std;

int CellCommunicator::getNextState(std::string currentState, unsigned int x, unsigned int y) const {
	if (_mode == MODE_RESET) {
		return _resetValue;
	}
	if (_mode == MODE_LOAD) { // Use hardcoded rules:
		if (currentState.size() == 9) {
			return (currentState[3]) - '0'; // state = W-node
		} else if (currentState.size() == 5) {
			return (currentState[1] - '0');
		} else if (currentState.size() == 3) {
			return (currentState[0] - '0');
		} else {
			std::cout << "Load mode not covered for neighbourness " << currentState << " with size " << currentState.size() << std::endl;
		}
		
	}
	if (_ruleTable) {
		return _ruleTable->getNextState(currentState, x, y);
	} else {
		std::cerr << "Rule table not set" << std::endl;
		return 0;
	}
}



Rules::Rules() {
}
// Normal operation:
void Rules::addRule(std::string key, int nextState) {
	_ruleTable[key] = nextState;
}

void Rules::addKeyValRule(std::string key, std::string value) {
	assert(value.size() == 8); // TODO: Refactor
	addRule(key + "111", value.at(0));
	addRule(key + "110", value.at(1));
	addRule(key + "101", value.at(2));
	addRule(key + "100", value.at(3));
	addRule(key + "011", value.at(4));
	addRule(key + "010", value.at(5));
	addRule(key + "001", value.at(6));
	addRule(key + "000", value.at(7));
}

int Rules::getNextState(std::string key, unsigned int x, unsigned int y) const {
	try {
		return _ruleTable.at(key);
	} catch (std::exception &e) {
		std::cout << "Key " << key << " exception ";
		std::cout << e.what() << std::endl;
		return -1;
	}
}

int Rules::getSize() const {
	return (int)_ruleTable.size();
}

std::vector<std::string> extendSingleRule(std::string rule, unsigned int to) {
	if (to - rule.size() <= 0) {
		cout << "Trying to reduce a rule, not implemented" << endl;
		assert(0);
	}
	size_t added = to - rule.size();
	std::vector<std::string> temp;
	generateAllPossibleStrings(temp, "", (added / 2) - 1);
	std::vector<std::string> result;
	for (unsigned int i = 0; i < temp.size(); i++) {
		for (unsigned int j = 0; j < temp.size(); j++) {
			result.push_back(temp[i] + rule + temp[j]);
		}
	}
	return result;
}

void Rules::extendToNeighbours(unsigned int neighbourness) {
	if (_ruleTable.begin()->first.size() == neighbourness) {
		return;
	}
	std::map<std::string, int> newRules;
	for (auto it = _ruleTable.begin(); it != _ruleTable.end(); ++it) {
		std::vector<std::string> extended = extendSingleRule(it->first, neighbourness);
		for (auto vecIt = extended.begin(); vecIt != extended.end(); ++vecIt) {
			newRules[*vecIt] = it->second;
		}
	}
	_ruleTable = newRules;
}

static void printVHDLSetRule(std::string prefix, std::string value) {
	std::cout << "\tOPCODE(31 downto 29) <= (OP_SETRULE);" << std::endl;
	if (prefix.length() > 0) {
		std::cout << "\tOPCODE(28 downto " << 28 - prefix.length() + 1 << ") <= \""
			<< prefix << "\";" << std::endl;
	}
	std::cout << "\tOPCODE(" << 28 - prefix.length() << " downto " << 28 - prefix.length() - 7
		<< ") <= \"" << value << "\";" << std::endl;
	std::cout << "\tWAIT UNTIL DONE = '1';" << std::endl;
}

static void printPseudoAssemblySetRule(std::string prefix, std::string value) {
	std::cout << "OP_SETRULE(";
	if (prefix.length() > 0) {
		std::cout << prefix << ",";
	}
	std::cout << value << ");";
}

void Rules::printAs(PrintModes printMode) const {
	// The degree of neighbourness
	int neighbourness = (_ruleTable.begin()->first).size();
	// The length of the leading bits
	int prefixLength = neighbourness - 3;
	// Prefix is the leading n bits
	std::string prefix = _ruleTable.begin()->first.substr(0, prefixLength);
	// The recombined rules
	std::vector<std::pair<std::string, int>> combinedRules;
	auto lastIt = _ruleTable.end();
	lastIt--;
	for (auto it = _ruleTable.begin(); it != _ruleTable.end(); ++it) {
		// The prefix has changed, store the rules for this prefix
		if (prefix != it->first.substr(0, prefixLength) || it == lastIt) {
			// Special case the very last rule:
			if (prefix == it->first.substr(0, prefixLength) && it == lastIt) {
				if (it->second == 1) {
					std::pair<std::string, int> rule(it->first, it->second);
					combinedRules.push_back(rule);
				}
			}
			// We only want to store if there is a atleast a single non-0 bit.
			int result = 0;
			std::string resultAsBin = "00000000";
			// Go through the collected rules, and combine them
			while (combinedRules.size()) {
				std::pair<std::string, int> rule = combinedRules.back();
				combinedRules.pop_back();
				string val = rule.first;
				assert(rule.second == 1);
				val.erase(0, val.size() - 3);
				int idx = 0;
				if (val.at(2) == '1') {
					idx += 1;
				}
				if (val.at(1) == '1') {
					idx += 2;
				}
				if (val.at(0) == '1') {
					idx += 4;
				}
				result += pow(2, idx);
				resultAsBin[7 - idx] = '1';
			}
			if (result != 0) {
				if (printMode == PRINT_VHDL) {
					printVHDLSetRule(prefix, resultAsBin);
				} else if (printMode == PRINT_PSEUDO_ASSEMBLY) {
					printPseudoAssemblySetRule(prefix, resultAsBin);
				}
			}
			prefix = it->first.substr(0, prefixLength);
		}
		if (it->second == 1) {
			std::pair<std::string, int> rule(it->first, it->second);
			combinedRules.push_back(rule);
		}
	}
}

void Rules::printAsVHDLTestBench() const {
	printAs(PRINT_VHDL);
}

void Rules::printAsPseudoAssembly() const {
	printAs(PRINT_PSEUDO_ASSEMBLY);
}
