#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <iomanip>
#include <cassert>
#include <stdexcept>
#include "utils.h"
#include "opcodes.h"
#include "system.h"
#include "rulesets.h"

using namespace std;

void runOperationInSimulator(Operation &op, System &testSystem) {
	switch (op.getOpcode()) {
		case OP_MAKEFUNC:
			// Assembly-only, fall through to OP_HALT.
		case OP_HALT:
			// Do nothing.
			break;
		case OP_RUN: {
			// TODO: Reuse the same code path as normal operation.
			int numGenerations = op.getIntArg(0);
			for (int i = 0; i < numGenerations; i++) {
				testSystem.updateSystem();
			}
			// If we have continuous printing this should happen.
			// testSystem.printSystem();
			break;
		}
		case OP_SETRULE: {
			// TODO: Support more than just 2 neighbour Wolfram
			// TODO: Allow a way to clear out any existing rules.
			Rules *rules = nullptr;
			if (testSystem.getNeighbourness() == 3) {
				string val = op.getArg(0);
				int ruleVal;
				if (val.length() == 8) {
					// Attempt to parse it as binary.
					ruleVal = bin2dec(val);
				} else {
					ruleVal = op.getIntArg(0);
				}
				rules = generateWolframRule(ruleVal);
				testSystem.setRuleSet(rules); // TODO: Fix leak
			} else {
				// TODO
				assert(0);
			}
			testSystem.setRuleSet(rules);
			break;
		}
		case OP_SETEDGE: { // [modeEW][25] [modeNS][24] [value][23 downto height - 1]
			bool wrapEW = op.getIntArg(0);
			bool wrapNS = op.getIntArg(1);
			// TODO: Actually set edge VALUE
			testSystem.setWrapEW(wrapEW);
			testSystem.setWrapNS(wrapNS);
			break;
		}
		case OP_LOADCOL: {
			int times = op.getIntArg(1);
			string value = op.getArg(0);
			// Special case: LoadCol should rotate the
			// system "times"-times if wrapping is enabled.
			if (testSystem.getWrapEW()) {
				for (int i = 0; i < times; i++) {
					testSystem.shiftSystemEastwards();
				}
			} else {
				int *column = new int[value.size()];
				for (int i = 0; i < value.size(); i++) {
					assert(isdigit(value.at(i)));
					// Note: convention is still uppermost is MSB
					// Thus we most store "backwards"
					column[value.size() - i - 1] = value.at(i) - '0';
				}
				for (int i = 0; i < times; i++) {
					testSystem.loadColumnIntoSystem(column, value.size());
				}
				delete[] column;
			}
			break;
		}
		case OP_RST:
			testSystem.resetSystemTo(op.getIntArg(0));
			// TODO: Reset rules too, for consistency with
			// the accelerator's behaviour.
			break;
		case OP_ASSERT: {
			std::string expected = op.getArg(0);
			std::string output = testSystem.getOutputColumn();
			bool match = (atoi(expected.c_str()) == bin2dec(output));
			cout << "Assert{" << endl;
			cout << "\tOutput: " << output << " : " << bin2dec(output) << endl;
			cout << "\tExpected: " << expected << endl;
			cout << "\tMatch: " << match << endl;
			cout << "}" << endl;
			assert(match == true);
			break;
		}
		case OP_PRINT: {
			std::string output = testSystem.getOutputColumn();
			cout << "Output: " << output << " : " << bin2dec(output) << endl;
			break;
		}
		// In simulator, this stores the values
		// which generate asserts in assembly.
		case OP_VERIFY_ALL: {
			// For consistnecy:
			testSystem.setWrapEW(true);
			testSystem.setWrapNS(true);
			std::string testVal = testSystem.getOutputColumn();
			for (int i = 0; i < testSystem.getWidth();i++) {
				std::string output = testSystem.getOutputColumn();
				testSystem.shiftSystemEastwards();
				op.pushResult(output);
			}
			if (testVal != testSystem.getOutputColumn()) {
				stringstream error;
				error << "ERROR: Resulting output-value is wrong after performing verification: \n"
					  << "\tExpected: " << testVal << endl
					  << "\tGot: " << testSystem.getOutputColumn() << endl;
				throw logic_error(error.str());
			}
			break;
		}
		case OP_VERIFY: {
			std::string output = testSystem.getOutputColumn();
			op.pushResult(output);
			break;
		}
		default:
			assert(0);
			break;
	}
}

void Program::runInSimulator() {
	// Verify that the settings are usable:
	if (_numNeighbours != 2 && _numNeighbours != 4 && _numNeighbours != 8) {
		stringstream ss;
		ss << "Num Neighbours " << _numNeighbours << " not handled by current implementation" << endl;
		throw logic_error(ss.str());
	}

	// TODO: Handle this on load.
	assert(getWidth() != -1);
	assert(getHeight() != -1);
	assert(getNumNeighbours() != -1);
	
	// TODO: Consider changing the neighbourhood size-notation to be the same
	// compared to the accelerator (2, 4, 8) vs (3, 5, 9).
	System testSystem(getWidth(), getHeight(), getNumNeighbours() + 1);
	// TODO: Allow the system to set printMode.
	//	testSystem.setPrintMode((printLst ? PRINTMODE_LST : PRINTMODE_TERM));
	// For now, we disable continuous printing, as we want
	// our output to be assembly only.
	testSystem.setPrintContinuously(false);

	try {
		for (auto it = _operations.begin(); it != _operations.end(); ++it) {
			runOperationInSimulator(*it, testSystem);
		}
	} catch (exception &e) {
		cerr << "Exception during simulation: " << e.what() << endl;
		exit(1);
	}
}
