#include <iostream>
#include <sstream>
#include <algorithm>
#include <cstring>
#include <cstdio>
#include <cassert>
#include <fstream>
#include "system.h"
#include "cell.h"
#include "rules.h"
#include "utils.h"

using namespace std;

System::System(unsigned int width, unsigned int height, unsigned int neighbourness, bool nonUniform) :
		_width(width),
		_height(height),
		_neighbourness(neighbourness),
		_wrapNS(false),
		_wrapEW(true),
		_unique(false),
		_printMode(PRINTMODE_TERM),
		_printContinuously(false) {
	_communicator = new CellCommunicator();
	_loadTable = new int[width * height];
	createCells();
	setupNeighbourhood(true);
}

System::~System() {
	destroyCells();
	delete _communicator;
	delete[] _loadTable;
}

void System::createCells() {
	_cells = new Cell**[_width];
	for (unsigned int x = 0; x < _width; x++) {
		_cells[x] = new Cell*[_height];
		for (unsigned int y = 0; y < _height; y++) {
			_cells[x][y] = new Cell(_neighbourness, _communicator, x, y);
		}
	}
}

void System::destroyCells() {
	for (unsigned int x = 0; x < _width; x++) {
		for (unsigned int y = 0; y < _height; y++) {
			delete _cells[x][y];
		}
		delete[] _cells[x];
	}
	delete[] _cells;
}

void System::resetSystemTo(int value) {
	RuleMode oldMode = _communicator->getMode();
	_communicator->setMode(MODE_RESET);
	_communicator->setResetValue(value);
	updateSystem();
	_communicator->setMode(oldMode);
	logHistory();
}

void System::loadColumnIntoSystem(const int *column, size_t height) {
	RuleMode oldMode = _communicator->getMode();
	_communicator->setMode(MODE_LOAD);
	bool wasWrapped = getWrapEW();
	setWrapEW(false);

	int *data = nullptr;
	if (_height > height) {
		data = new int[_height];
		memcpy(data, column, sizeof(unsigned int) * height);
		memset(data + height, 0, (_height - height) * sizeof(unsigned int));
		height = _height;
	} else if (_height < height) {
		height = _height;
		data  = new int [_height];
		memcpy(data, column, sizeof(unsigned int) * height);
	} else {
		data = new int[_height];
		memcpy(data, column, sizeof(unsigned int) * height);
		height = _height;
	}

	Cell leftTemp(0, _communicator, 0, 0);
	
	int leftNeighbourIndex = 0;
	if (_neighbourness == 9) {
		leftNeighbourIndex = 3;//_neighbourness / 2 - 2;
	} else if (_neighbourness == 5) {
		leftNeighbourIndex = 1;
	} else if (_neighbourness == 3) {
		leftNeighbourIndex = 0;
	} else {
		cout << "loadColumnIntoSystem - unknown neighbourness: " << _neighbourness << endl;
		assert(0);
	}

	
	// Connect the left-side to our mem-buffer:
	Cell **lhs = new Cell*[_height];
	const Cell **oldNeighbours = new const Cell*[_height];
	for (unsigned int i = 0; i < _height; i++) {
		lhs[i] = new Cell(0, _communicator, 0, 0);
		oldNeighbours[i] = _cells[_width - 1][i]->getNeighbour(leftNeighbourIndex);
		_cells[_width - 1][i]->setNeighbour(leftNeighbourIndex, lhs[i]);
	}
	
	for (unsigned int y = 0; y < _height; y++) {
		lhs[y]->setState(data[y]);
	}

	updateSystem();

	for (unsigned int i = 0; i < height; i++) {
		_cells[0][i]->setNeighbour(leftNeighbourIndex, oldNeighbours[i]);
		delete lhs[i];
	}
	delete[] lhs;
	delete[] oldNeighbours;
	delete[] data;
	_communicator->setMode(oldMode);
	setWrapEW(wasWrapped);
}

void System::loadArrayIntoSystem(const int *array, unsigned int width, unsigned int height) {
	memcpy(_loadTable, array, sizeof(int) * std::min(_width * _height, width * height));

	int counter = 0;

	int *tempTable = new int[height];
	for (int x = _width - 1; x >= 0; x--) { // We want the right hand side of the table first
		for (unsigned int y = 0; y < _height; y++) {
			tempTable[y] = _loadTable[get1DIndex(x, y, _width)];
		}
		loadColumnIntoSystem(tempTable, height);
		counter++;
	}
	delete[] tempTable;
}

std::string trimString(std::string input) {
	std::string result;
	for (unsigned int i = 0; i < input.size(); i++) {
		if (input[i] >= '0' && input[i] <= '9') {
			result += input[i];
		}
	}
	return result;
}

/**
 * Load a text file into the system
 *
 * The text file should contain one line per row, 
 * and enough elements in each row to fill the row.
 * @param filename the filename of the file to be loaded.
 */
void System::loadFileIntoSystem(const std::string &filename) {
	std::ifstream file;
	file.open(filename);
	if (!file.good()) {
		std::cerr << "Unable to open file " << filename << std::endl;
		exit(1);
	}

	vector<string> lines;
	while (!file.eof()) {
		string line;
		getline(file, line);
		line = trimString(line);
		// Ignore any empty lines.
		if (line.size())
			lines.push_back(line);
	}
	// Empty file, no point in continuing.
	if (lines.size() == 0) {
		return;
	}
	size_t minLength = lines[0].size();
	for (unsigned int i = 0; i < lines.size(); i++) {
		minLength = min(minLength, lines[i].size());
	}

	int *table = new int[lines.size()];
	for (size_t x = minLength; x > 0; x--) {
		for (size_t y = 0; y < lines.size(); y++) { // Top line is highest coord internally
			char convert[2] = {0}; // Allow room for \0
			convert[0] = lines[y].at(x - 1);
			table[lines.size() - y - 1] = atoi(convert); // Assign flipped, so coords are correct.
		}
		loadColumnIntoSystem(table, lines.size());
	}
	delete[] table;
}

std::string System::getOutputColumn() {
	std::vector<string> historyItem;
	std::string column = "";
	for (int y = _height - 1; y >= 0; y--) {
		column += _cells[0][y]->getStateAsString();
	}
	return column;
}

void System::shiftSystemEastwards() {
	RuleMode oldMode = _communicator->getMode();
	_communicator->setMode(MODE_LOAD);
	updateSystem();
	_communicator->setMode(oldMode);
}

void System::setWrapEW(bool value) {
	_wrapEW = value;
	setupNeighbourhood(false);
}

void System::setWrapNS(bool value) {
	_wrapNS = value;
	setupNeighbourhood(false);
}

void System::outputStateAsPseudoAssembly(bool includeRules) {
	cout << "/* PseudoAssembly TestBench dump from C++ CA-simulator*/" << endl;
	if (includeRules) {
		cout << "/* Rule dump - init:*/" << endl;
		cout << "#height(" << _height << ");" << endl
			 << "#neighbours(" << _neighbourness - 1 << ");" << endl
			 << "#width("<< _width << ");" << endl
			 << "OP_RST(0);" << endl
			 << "OP_SETEDGE(0,0,0);" << endl
			 << "/* Starting rule-dump*/" << endl;
		_communicator->getRuleSet()->printAsPseudoAssembly();
	}
	bool wrapState = getWrapEW();
	string prevColState = getOutputColumn();
	int count = 0;
	setWrapEW(true);
	cout << "/* Start of state dump*/" << endl;
	stringstream ss;
	string stateRange = ss.str();
	for (int i = 0; i < getWidth(); i++) {
		string curColState = getOutputColumn();
		if (curColState != prevColState) {
			cout << "OP_LOADCOL(" << prevColState << "," << count << ");" << endl;
			count = 1;
		} else {
			count++;
		}
		prevColState = curColState;
		shiftSystemEastwards();
	}
	if (count != 0) {
		cout << "OP_LOADCOL(" << prevColState << "," << count << ");" << endl;
	}
	cout << "/* End of state dump */" << endl;
	setWrapEW(wrapState);
}


void System::outputStateAsVHDLTB(bool includeRules) {
	cout << "-- VHDL TestBench dump from C++ CA-simulator" << endl;
	if (includeRules) {
		cout << "-- Rule dump - init:" << endl;
		cout << "\tWAIT FOR 100 ns;" << endl; // Allow the testbench to settle
		cout << "\tOPCODE(31 downto 29) <= (OP_RST);" << endl;
		cout << "\tOPCODE(28) <= '0';" << endl; // We have a hardcoded "0" for initialization.
		cout << "\tWAIT UNTIL DONE = '1';" << endl;
		cout << "-- Starting rule-dump" << endl;
		_communicator->getRuleSet()->printAsVHDLTestBench();	
	}
	bool wrapState = getWrapEW();
	string prevColState = getOutputColumn();
	int count = 0;
	setWrapEW(true);
	cout << "-- Start of state dump" << endl;
	stringstream ss;
	ss << "28";
	ss << " downto " << 28 - getHeight() + 1;
	ss << ")";
	string stateRange = ss.str();
	for (int i = 0; i < getWidth(); i++) {
		string curColState = getOutputColumn();
		if (curColState != prevColState) {
			cout << "\tOPCODE(31 downto 29) <= (OP_LOADCOL);" << endl;
			cout << "\tOPCODE(" << stateRange << " <= \"" << prevColState << "\";" << endl;
			cout << "\tOPCODE(28 - " << getHeight() << " downto 28 - " << getHeight() << " - 7) <= conv_std_logic_vector(" << count << ",8);" << endl;
			cout << "\tWAIT UNTIL DONE = '1';" << endl << endl;
			count = 1;
		} else {
			count++;
		}
		prevColState = curColState;
		shiftSystemEastwards();
	}
	if (count != 0) {
		cout << "\tOPCODE(31 downto 29) <= (OP_LOADCOL);" << endl;
		cout << "\tOPCODE(" << stateRange << " <= \"" << prevColState << "\";" << endl;
		cout << "\tOPCODE(28 - " << getHeight() << " downto 28 - " << getHeight() << " - 7) <= conv_std_logic_vector(" << count << ",8);" << endl;
		cout << "\tWAIT UNTIL DONE = '1';" << endl;
	}
	cout << "-- End of state dump" << endl;
	setWrapEW(wrapState);
}

void System::setupNeighbourhood(bool resetState) {
	int NW, SW, N, W, C, E, S, NE, SE;
	C = (_neighbourness / 2);
	NW = C - 4;
	SW = C - 3;
	N =  C - 2;
	W =  C - 1;
	E =  C + 1;
	S =  C + 2;
	NE = C + 3;
	SE = C + 4;
	if (_neighbourness != 3 && _neighbourness != 5 && _neighbourness != 9) {
		assert(0);
	}

	for (unsigned int y = 0; y < _height; y++) {
		for (unsigned int x = 0; x < _width; x++) {
			if (resetState) {
				_cells[x][y]->setState(0);
			}
			_cells[x][y]->clearAllNeighbours();
			_cells[x][y]->setNeighbour(C, _cells[x][y]);
			_cells[x][y]->setNumNeighbours(_neighbourness);
			if (_neighbourness >= 3) {
				if (x > 0) { // E
					_cells[x][y]->setNeighbour(E, _cells[x - 1][y]);
				} else if (x == 0 && _wrapEW) {
					_cells[x][y]->setNeighbour(E, _cells[_width - 1][y]);
				}
				if (x < (_width - 1)) { // W
					_cells[x][y]->setNeighbour(W, _cells[x + 1][y]);
				} else if (x == (_width - 1) && _wrapEW) {
					_cells[x][y]->setNeighbour(W, _cells[0][y]);
				}
			} 
			if (_neighbourness >= 5) {
				// N, W, c, E, S
				if (y > 0) { // S
					_cells[x][y]->setNeighbour(S, _cells[x][y - 1]);
				} else if (y == 0 && _wrapNS) {
					_cells[x][y]->setNeighbour(S, _cells[x][_height - 1]);
				}
				if (y < (_height - 1)) { // N
					_cells[x][y]->setNeighbour(N, _cells[x][y + 1]);
				} else if (y == (_height - 1) && _wrapNS) {
					_cells[x][y]->setNeighbour(N, _cells[x][0]);
				}
			} 
			if (_neighbourness == 9) {
				// NW, SW, N, W, c, E, S, NE, SE
				if (y > 0) { // S
					if (x > 0) { // SE
						_cells[x][y]->setNeighbour(SE, _cells[x - 1][y - 1]);
					} else if (x == 0 && _wrapEW) {
						_cells[x][y]->setNeighbour(SE, _cells[_width - 1][y - 1]);
					}
					if (x < (_width - 1)) { // SW
						_cells[x][y]->setNeighbour(SW, _cells[x +  1][y - 1]);
					} else if (x == (_width - 1) && _wrapEW) {
						_cells[x][y]->setNeighbour(SW, _cells[0][y - 1]);
					}
				} else if (y == 0 && _wrapNS) {
					if (x > 0) { // SE
						_cells[x][y]->setNeighbour(SE, _cells[x - 1][_height - 1]);
					} else if (x == 0 && _wrapEW) {
						_cells[x][y]->setNeighbour(SE, _cells[_width - 1][_height - 1]); // Wraps from NW
					}
					if (x < (_width - 1)) { // SW
						_cells[x][y]->setNeighbour(SW, _cells[x +  1][_height - 1]);
					} else if (x == _width - 1 && _wrapEW) {
						_cells[x][y]->setNeighbour(SW, _cells[0][_height - 1]); // Wraps from NE
					}
				}
				if (y < (_height - 1)) { // N
					if (x > 0) { // NE
						_cells[x][y]->setNeighbour(NE, _cells[x - 1][y + 1]);
					} else if (x == 0 && _wrapEW) {
						_cells[x][y]->setNeighbour(NE, _cells[_width - 1][y + 1]);
					}
					if (x < (_width - 1)) { // NW
						_cells[x][y]->setNeighbour(NW, _cells[x +  1][y + 1]);
					} else if (x == (_width - 1)) {
						_cells[x][y]->setNeighbour(NW, _cells[0][y + 1]);
					}
				} else if (y == (_height - 1) && _wrapNS) {
					if (x > 0) { // NE
						_cells[x][y]->setNeighbour(NE, _cells[x - 1][0]);
					} else if (x == 0 && _wrapEW) {
						_cells[x][y]->setNeighbour(NE, _cells[_width - 1][0]); // wraps from SW
					}
					if (x < (_width - 1)) { // NW
						_cells[x][y]->setNeighbour(NW, _cells[x +  1][0]);
					} else if (x == _width - 1 && _wrapEW) {
						_cells[x][y]->setNeighbour(NW, _cells[0][0]); // wraps from SE
					}
				}
			}
			if (_neighbourness != 3 && _neighbourness != 5 && _neighbourness != 9) {
				std::cout << "Unhandled neighbourness " << _neighbourness << std::endl;
			}
		}
	}
}

void System::setRuleSet(const Rules *ruleset) {
	_communicator->setRuleSet(ruleset);
}

void System::printSystem() const {
	for (unsigned int y = _height; y > 0; y--) {
		for (unsigned int x = _width; x > 0; x--) {
			std::cout << _cells[x - 1][y - 1]->getStateAsString();
		}
		std::cout << std::endl;
	}
}
void System::updateSystem() { // Update west to east
	for (unsigned int y = _height; y > 0; y--) {
		for (unsigned int x = _width; x > 0; x--) {
			_cells[x - 1][y - 1]->update();
		}
	}
	
	for (unsigned int x = _width; x > 0; x--) {
		for (unsigned int y = _height; y > 0; y--) {
			_cells[x - 1][y - 1]->tick();
		}
	}
	logHistory();
}

void System::logHistory() {
	std::vector<string> historyItem;
	for (int y = _height; y > 0; y--) {
		string line = "";
		for (unsigned int x = _width; x > 0; x--) {
			line += _cells[x - 1][y - 1]->getStateAsString();
		}
		historyItem.push_back(line);
	}
	_history.push_back(historyItem);
	if (_printContinuously) {
		printSystem();
		cout << endl;
	}
}

void System::clearHistory() {
	std::vector<std::string> lastItem = _history.back();
	_history.clear();
	_history.push_back(lastItem);
}

void System::printHistory() const {
	for (unsigned int i = 0; i < _history.size(); i++) {
		bool isBlank = true;
		for (unsigned int j = 0; j < _history[i].size(); j++) {
			if (_printMode == PRINTMODE_LST) {
				if (!_unique || i == 0 || (i > 0 && _history[i][j] != _history[i - 1][j])) {
					isBlank = false;
					if (j == 0) {
						cout << i << "\t" << j << "\t";
					}
					cout << " " << _history[i][j];
				}
			} else if (_printMode == PRINTMODE_TERM) {
				isBlank = false;
				cout << _history[i][j] << endl;
			}
		}
		if (!isBlank) {
			cout << endl;
		}
	}
}
